#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utils for data_gen"""

import enum
from abc import ABC
from typing import Dict, List, Text, Any

import apache_beam as beam
import tensorflow as tf

from zenml.logger import get_logger

logger = get_logger(__name__)


class DataType(enum.IntEnum):
    UNKNOWN = -1
    INT = 0
    FLOAT = 1
    BYTES = 2


def _int_converter(value):
    if value is None or value is '' or value is b'':
        return []
    else:
        return [int(value)]


def _float_converter(value):
    if value is None or value is '' or value is b'':
        return []
    else:
        return [float(value)]


def _bytes_converter(value):
    if value is None:
        return []
    else:
        return [tf.compat.as_bytes(value)]


SCHEMA_MAPPING = {
    'INTEGER': DataType.INT,
    'FLOAT': DataType.FLOAT,
    'STRING': DataType.BYTES,
    'BYTES': DataType.BYTES,
    'TIMESTAMP': DataType.BYTES,
    'DATETIME': DataType.BYTES
}

CONVERTER_MAPPING = {
    DataType.UNKNOWN: lambda x: [],
    DataType.INT: _int_converter,
    DataType.FLOAT: _float_converter,
    DataType.BYTES: _bytes_converter,
}


@beam.typehints.with_output_types(beam.typehints.Dict[Text, DataType])
class DtypeInferrer(beam.CombineFn, ABC):
    """A beam.CombineFn to infer data types"""

    @staticmethod
    def _infer_value_type(value) -> DataType:
        try:
            float(value)
        except ValueError:
            if not value:
                return DataType.UNKNOWN

            elif type(value) is bytes or type(value) is str:
                return DataType.BYTES
            else:
                return DataType.UNKNOWN

        except TypeError:
            return DataType.UNKNOWN
        else:
            if float(value).is_integer():
                # There is still a case that it won't be an integer, if its
                # 0.0, 1.0, 2.0. So we try and convert explicitly here
                try:
                    int(value)
                except ValueError:
                    return DataType.FLOAT
                return DataType.INT
            else:
                return DataType.FLOAT

    def create_accumulator(self, **kwargs) -> Dict[Text, DataType]:
        return {}

    def add_input(self,
                  accumulator: Dict[Text, DataType],
                  element: Dict[Text, bytes],
                  **kwargs,
                  ) -> Dict[Text, DataType]:

        for key, value in element.items():
            previous_type = accumulator.get(key, None)
            current_type = self._infer_value_type(value)

            if previous_type is None or current_type > previous_type:
                accumulator[key] = current_type
        return accumulator

    def merge_accumulators(self,
                           accumulators: List[Dict[Text, DataType]],
                           **kwargs,
                           ) -> Dict[Text, DataType]:

        result = {}
        for single_accumulator in accumulators:
            for feature_name, feature_type in single_accumulator.items():
                if feature_name not in result \
                        or feature_type > result[feature_name]:
                    result[feature_name] = feature_type
        return result

    def extract_output(self,
                       accumulator: Dict[Text, Any],
                       **kwargs) -> Dict[Text, Text]:
        for k in accumulator:
            if k == DataType.UNKNOWN:
                logger.warning(f'Unknown data type in the schema '
                               f'for feature {k}!')
        return accumulator


def append_tf_example(data: Dict[Text, Any],
                      schema: Dict[Text, Any]) -> tf.train.Example:
    """Add tf example to row"""
    feature = {}
    for key, value in data.items():
        data_type = schema[key]
        value = CONVERTER_MAPPING[data_type](value)
        if data_type == DataType.INT:
            feature[key] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=value))
        elif data_type == DataType.FLOAT:
            feature[key] = tf.train.Feature(
                float_list=tf.train.FloatList(value=value))
        elif data_type == DataType.BYTES:
            feature[key] = tf.train.Feature(
                bytes_list=tf.train.BytesList(value=value))
        else:
            feature[key] = tf.train.Feature(
                bytes_list=tf.train.BytesList(value=value))

    tf_example = tf.train.Example(features=tf.train.Features(feature=feature))
    return tf_example



import os

from zenml.backends.orchestrator import OrchestratorKubernetesBackend
from zenml.datasources import CSVDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.metadata import MySQLMetadataStore
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository, ArtifactStore
from zenml.steps.evaluator import TFMAEvaluator
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.utils.naming_utils import transformed_label_name

GCP_PROJECT = os.getenv('GCP_PROJECT')
GCP_BUCKET = os.getenv('GCP_BUCKET')
GCP_REGION = os.getenv('GCP_REGION')
GCP_CLOUD_SQL_INSTANCE_NAME = os.getenv('GCP_CLOUD_SQL_INSTANCE_NAME')
MYSQL_DB = os.getenv('MYSQL_DB')
MYSQL_USER = os.getenv('MYSQL_USER')
MYSQL_PWD = os.getenv('MYSQL_PWD')
MYSQL_HOST = os.getenv('MYSQL_HOST', '127.0.0.1')
MYSQL_PORT = os.getenv('MYSQL_PORT', 3306)
CONNECTION_NAME = f'{GCP_PROJECT}:{GCP_REGION}:{GCP_CLOUD_SQL_INSTANCE_NAME}'
# Path to your kubernetes config:
K8S_CONFIG_PATH = os.path.join(os.environ["HOME"], '.kube/config')

assert GCP_BUCKET
assert GCP_PROJECT
assert GCP_REGION
assert MYSQL_DB
assert MYSQL_USER
assert MYSQL_PWD

# Run the pipeline on a kubernetes cluster.
# The metadata store and artifact store should be accessible by the cluster.

# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = CSVDatasource(name='Pima Indians Diabetes',
                       path='gs://zenml_quickstart/diabetes.csv')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'Pima Indians Diabetes')
training_pipeline.add_datasource(ds)

training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    StandardPreprocesser(
        features=['times_pregnant', 'pgc', 'dbp', 'tst', 'insulin', 'bmi',
                  'pedigree', 'age'],
        labels=['has_diabetes'],
        overwrite={'has_diabetes': {
            'transform': [{'method': 'no_transform', 'parameters': {}}]}}
    ))
# Add a trainer
training_pipeline.add_trainer(TFFeedForwardTrainer(
    loss='binary_crossentropy',
    last_activation='sigmoid',
    output_units=1,
    metrics=['accuracy'],
    epochs=20))

# Add an evaluator
training_pipeline.add_evaluator(
    TFMAEvaluator(slices=[['has_diabetes']],
                  metrics={transformed_label_name('has_diabetes'):
                               ['binary_crossentropy', 'binary_accuracy']}))

# Define the metadata store
metadata_store = MySQLMetadataStore(
    host=MYSQL_HOST,
    port=int(MYSQL_PORT),
    database=MYSQL_DB,
    username=MYSQL_USER,
    password=MYSQL_PWD,
)

# Define the artifact store
artifact_store = ArtifactStore(
    os.path.join(GCP_BUCKET, 'kubernetes_orcestrated/artifact_store'))

# Define the orchestrator backend
orchestrator_backend = OrchestratorKubernetesBackend(
    kubernetes_config_path=K8S_CONFIG_PATH,
    image_pull_policy="Always")

# Run the pipeline on a Kubernetes Cluster
training_pipeline.run(
    backend=orchestrator_backend,
    metadata_store=metadata_store,
    artifact_store=artifact_store,
)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from abc import abstractmethod

from tensorflow_metadata.proto.v0.schema_pb2 import Schema
from tensorflow_metadata.proto.v0.statistics_pb2 import \
    DatasetFeatureStatisticsList

from zenml.steps import BaseStep
from zenml.enums import StepTypes


class BaseSequencerStep(BaseStep):
    """
    Base class for all sequencer steps. These steps are used to
    specify transformation and filling operations on timeseries datasets
    that occur before the data preprocessing takes place.
    """

    STEP_TYPE = StepTypes.sequencer.name

    def __init__(self,
                 statistics: DatasetFeatureStatisticsList = None,
                 schema: Schema = None,
                 **kwargs):
        """
        Base Sequencer constructor.

        This steps uses a beam pipeline to handle the data processing and the
        pipeline consists of a few steps, where the main logic for each
        datapoint can be summarized as follows:

            1 - Add a timestamp to the datapoint
            2 - Add a category key to the datapoint (optional)
            3 - Split your data into sessions based on a windowing strategy
            4 - Process the sessions to create sequences

        With the `abstractmethod`s listed below, you have the option to modify
        anyone of these steps.

        Args:
            statistics: Parsed statistics output of a preceding StatisticsGen.
            schema: Parsed schema output of a preceding SchemaGen.
        """

        super().__init__(**kwargs)

        self.statistics = statistics
        self.schema = schema

    @abstractmethod
    def get_timestamp_do_fn(self):
        """
        The process of sequencing is highly dependent on the format of your
        data. For instance, the timestamp of a single datapoint can be
        infused within the datapoint in various shapes or forms.

        It is impossible to find THE one solution which would be able to parse
        all the different variants of timestamps and that is why we exposed
        this method.

        Through this method, you have access to all of the instance
        variables of your step and all you have to do is to return an instance
        of a beam.DoFn which returns a TimestampedValue. You can check our
        StandardSequencer for a practical example.
        """
        pass

    @abstractmethod
    def get_category_do_fn(self):
        """
        In ZenML, you have the option to split your data based on a categorical
        feature before the actual sequencing happens. This is especially
        helpful if you are dealing with a joint dataset (i.e dataset featuring
        multiple assets in the field, but you want to sequence on an
        asset-level)

        Similar to get_timestamp_do_fn, you need to implement a method, which
        returns an instance of a beam.DoFn class. This beam.DoFn should be
        responsible for extracting the category of a datapoint and add it to
        the datapoint and return it. For a practical example, you can check our
        StandardSequencer.
        """
        # TODO: Implement a default method, in case categories are not used
        pass

    @abstractmethod
    def get_window(self):
        """
        This method needs to return the desired windowing strategy for the
        beam pipeline.
        """
        pass

    @abstractmethod
    def get_combine_fn(self):
        """
        Once the data is split into sessions (and possibly categories too),
        it needs to be processed in order to extract sequences from the
        sessions.

        This method needs to return an instance of beam.CombineFn class, which
        processes the accumulated datapoints and extracts desired sequences.
        You can check out our StandardSequencer for a practical example.
        """
        pass



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import importlib.util
import os
from typing import Dict, Text, List

from zenml.components.pusher import cortex_executor
from zenml.repo import Repository
from zenml.steps.deployer import BaseDeployerStep
from zenml.logger import get_logger
from zenml.utils.source_utils import get_path_from_source, \
    get_class_source_from_source, resolve_class

spec = importlib.util.find_spec('cortex')
if spec is None:
    raise AssertionError("cortex integration not installed. Please install "
                         "zenml[cortex] via `pip install zenml[cortex]`")

logger = get_logger(__name__)


class CortexDeployer(BaseDeployerStep):
    """
    Step class for deploying models using Cortex.
    """

    def __init__(self,
                 api_config: Dict,
                 predictor=None,
                 requirements: List = None,
                 conda_packages: List = None,
                 env: Text = 'gcp',
                 force: bool = True,
                 wait: bool = False,
                 **kwargs):
        """
        Cortex Deployer Step constructor.

        Use this step to push your trained model via the
        [Cortex API](https://docs.cortex.dev/).

        Args:
            api_spec: A dictionary defining a single Cortex API. See
            https://docs.cortex.dev/v/0.27/ for schema.
            predictor: A Cortex Predictor class implementation. Not required
            for TaskAPI/TrafficSplitter kinds.
            requirements: A list of PyPI dependencies that will be installed
            before the predictor class implementation is invoked.
            conda_packages: A list of Conda dependencies that will be
            installed before the predictor class implementation is invoked.
            force: Override any in-progress api updates.
            wait: Streams logs until the APIs are ready.
        """
        if conda_packages is None:
            conda_packages = []
        if requirements is None:
            requirements = ['tensorflow==2.3.0']

        self.env = env
        self.predictor = predictor
        self.api_config = api_config
        self.requirements = requirements
        self.conda_packages = conda_packages
        self.force = force
        self.wait = wait
        super(CortexDeployer, self).__init__(
            env=self.env,
            predictor=predictor,
            api_config=self.api_config,
            requirements=self.requirements,
            conda_packages=self.conda_packages,
            force=self.force,
            wait=self.wait,
            **kwargs)

    def get_config(self):
        predictor_path = self.predictor.__module__ + '.' + \
                         self.predictor.__name__
        # TODO: [MEDIUM] Find a better way to get p_file_path
        p_file_path = \
            get_path_from_source(get_class_source_from_source(predictor_path))
        repo: Repository = Repository.get_instance()

        return {
            "cortex_serving_args": {
                "env": self.env,
                "api_config": self.api_config,
                "predictor_path": os.path.join(repo.path, p_file_path),
                "requirements": self.requirements,
                "conda_packages": self.conda_packages,
                "force": self.force,
                "wait": self.wait,
            }}

    def get_executor(self):
        return cortex_executor.Executor



from typing import Any, Dict, List, Text

import attr
from tfx import types
from tfx.components.trainer import fn_args_utils
from tfx.components.trainer.executor import GenericExecutor
from tfx.components.trainer.fn_args_utils import FnArgs
from tfx.types import artifact_utils
from tfx.utils import io_utils
from tfx.utils import json_utils
from tfx.utils import path_utils

from zenml.components.trainer import constants
from zenml.logger import get_logger

logger = get_logger(__name__)

_TELEMETRY_DESCRIPTORS = ['Trainer']


class ZenMLTrainerArgs(FnArgs):
    input_patterns = attr.ib(type=Dict[Text, Text], default=None)
    output_patterns = attr.ib(type=Dict[Text, Text], default=None)


class ZenMLTrainerExecutor(GenericExecutor):
    """
    An adjusted version of the TFX Trainer Generic executor, which also
    handles an additional output artifact while managing the fn_args
    """

    def _GetFnArgs(self,
                   input_dict: Dict[Text, List[types.Artifact]],
                   output_dict: Dict[Text, List[types.Artifact]],
                   exec_properties: Dict[Text, Any]) -> fn_args_utils.FnArgs:

        result = ZenMLTrainerArgs()

        # INPUTS ##############################################################
        if input_dict.get(constants.TRANSFORM_GRAPH):
            transform_graph_path = artifact_utils.get_single_uri(
                input_dict[constants.TRANSFORM_GRAPH])
        else:
            transform_graph_path = None

        if input_dict.get(constants.SCHEMA):
            schema_path = io_utils.get_only_uri_in_dir(
                artifact_utils.get_single_uri(
                    input_dict[constants.SCHEMA]))
        else:
            schema_path = None

        input_examples = input_dict[constants.EXAMPLES]
        splits = artifact_utils.decode_split_names(
            artifact_utils.get_single_instance(input_examples).split_names)

        input_patterns = dict()
        for split in splits:
            input_patterns.update({
                split: io_utils.all_files_pattern(uri) for uri in
                artifact_utils.get_split_uris(input_examples, split)})

        result.schema_path = schema_path
        result.transform_output = transform_graph_path
        result.input_patterns = input_patterns

        # PARAMETERS ##########################################################
        custom_config = json_utils.loads(
            exec_properties.get(constants.CUSTOM_CONFIG, 'null'))
        result.custom_config = custom_config

        # OUTPUTS #############################################################
        output_artifact = output_dict[constants.TEST_RESULTS]
        output_instance = artifact_utils.get_single_instance(output_artifact)
        output_instance.split_names = artifact_utils.encode_split_names(splits)
        output_patterns = dict()
        for split in splits:
            output_patterns.update({
                split: uri for uri in
                artifact_utils.get_split_uris(output_artifact, split)})

        out_path = artifact_utils.get_single_uri(output_dict[constants.MODEL])
        serving_model_dir = path_utils.serving_model_dir(out_path)

        result.output_patterns = output_patterns
        result.serving_model_dir = serving_model_dir

        return result



from preprocessing.preprocessor import ImagePreprocessor
from trainer.trainer_step import ImageTensorflowTrainer
from zenml.datasources.tfds_datasource import TFDSDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository
from zenml.steps.evaluator.tfma_evaluator import TFMAEvaluator
from zenml.steps.split.random_split import RandomSplit
from zenml.utils import naming_utils

# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = TFDSDatasource(name='MNIST', tfds_dataset_name='mnist')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name('MNIST')
training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    ImagePreprocessor(
        image_key='image',
        labels=['label'],
        shape=[-1, 28, 28, 3]
    ))

# Add a trainer
training_pipeline.add_trainer(ImageTensorflowTrainer(
    loss='binary_crossentropy',
    last_activation='sigmoid',
    metrics=['accuracy'],
    epochs=5))

# Add an evaluator
training_pipeline.add_evaluator(
    TFMAEvaluator(slices=[[naming_utils.transformed_label_name('label')]],
                  metrics={naming_utils.transformed_label_name('label'): [
                      'binary_accuracy']}))

# Run the pipeline locally
training_pipeline.run()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


class ConfigKeys:
    @classmethod
    def get_keys(cls):
        keys = {key: value for key, value in cls.__dict__.items() if
                not isinstance(value, classmethod) and
                not isinstance(value, staticmethod) and
                not callable(value) and
                not key.startswith('__')}

        required = {k: v for k, v in keys.items() if not k.endswith('_')}
        optional = {k: v for k, v in keys.items() if k.endswith('_')}

        return required, optional

    @classmethod
    def key_check(cls, _input):
        # Check whether the given config is a dict
        """
        Args:
            _input:
        """
        assert isinstance(_input, dict), 'Please specify a dict for {}' \
            .format(cls.__name__)

        # Required and optional keys for the config dict
        required, optional = cls.get_keys()

        # Check for missing keys
        missing_keys = [k for k in required.values() if
                        k not in _input.keys()]
        assert len(missing_keys) == 0, \
            'Missing key(s) {} in {}'.format(missing_keys,
                                             cls.__name__)

        # Check for unknown keys
        unknown_keys = [k for k in _input.keys() if
                        k not in required.values() and
                        k not in optional.values()]
        assert len(unknown_keys) == 0, \
            'Unknown key(s) {} in {}. Required keys : {} ' \
            'Optional Keys: {}'.format(unknown_keys,
                                       cls.__name__,
                                       list(required.values()),
                                       list(optional.values()))


class GlobalKeys(ConfigKeys):
    VERSION = 'version'
    ARTIFACT_STORE = 'artifact_store'
    METADATA_STORE = 'metadata'
    BACKEND = 'backend'
    PIPELINE = 'pipeline'


class PipelineKeys(ConfigKeys):
    ARGS = 'args'
    SOURCE = 'source'
    STEPS = 'steps'
    DATASOURCE = 'datasource'
    DATASOURCE_COMMIT_ID = 'datasource_commit_id'
    ENABLE_CACHE = 'enable_cache'


class PipelineDetailKeys(ConfigKeys):
    NAME = 'name'
    TYPE = 'type'
    ENABLE_CACHE = 'enable_cache'


class DatasourceKeys(ConfigKeys):
    ID = 'id'
    NAME = 'name'
    SOURCE = 'source'
    ARGS = 'args'
    METADATA_STORE = 'metadata'
    ARTIFACT_STORE = 'artifact_store'
    BACKEND = 'backend'


class BackendKeys(ConfigKeys):
    TYPE = 'type'
    ARGS = 'args'
    SOURCE = 'source'


class MLMetadataKeys(ConfigKeys):
    TYPE = 'type'
    ARGS = 'args'


class StepKeys(ConfigKeys):
    NAME = 'name'
    SOURCE = 'source'
    ARGS = 'args'
    BACKEND = 'backend'


class TrainingSteps(ConfigKeys):
    DATA = 'data'
    SPLIT = 'split'
    SEQUENCER = 'sequencer'
    PREPROCESSER = 'preprocesser'
    TRAINER = 'trainer'
    EVALUATOR = 'evaluator'
    DEPLOYER = 'deployer'


class NLPSteps(ConfigKeys):
    SPLIT = 'split'
    TRAINER = 'trainer'
    TOKENIZER = 'tokenizer'


class DataSteps(ConfigKeys):
    DATA = 'data'


class InferSteps(ConfigKeys):
    DATA = 'data'
    INFER = 'infer'


# HELPER KEYS

class MethodKeys(ConfigKeys):
    METHOD = 'method'
    PARAMETERS = 'parameters'


class DefaultKeys(ConfigKeys):
    STRING = 'string'
    INTEGER = 'integer'
    BOOLEAN = 'boolean'
    FLOAT = 'float'



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


import click
from dateutil import tz

from zenml.repo import Repository, GlobalConfig

pass_config = click.make_pass_decorator(GlobalConfig, ensure=True)

pass_repo = click.make_pass_decorator(Repository, ensure=True)


def title(text):
    """
    Args:
        text:
    """
    click.echo(click.style(text.upper(), fg='cyan', bold=True, underline=True))


def confirmation(text, *args, **kwargs):
    """
    Args:
        text:
        *args:
        **kwargs:
    """
    return click.confirm(click.style(text, fg='yellow'), *args,
                         **kwargs)


def question(text, *args, **kwargs):
    """
    Args:
        text:
        *args:
        **kwargs:
    """
    return click.prompt(text=text, *args, **kwargs)


def declare(text):
    """
    Args:
        text:
    """
    click.echo(click.style(text, fg='green'))


def notice(text):
    """
    Args:
        text:
    """
    click.echo(click.style(text, fg='cyan'))


def error(text):
    """
    Args:
        text:
    """
    raise click.ClickException(message=click.style(text, fg='red', bold=True))


def warning(text):
    """
    Args:
        text:
    """
    click.echo(click.style(text, fg='yellow', bold=True))


def pretty_print(obj):
    """
    Args:
        obj:
    """
    click.echo(str(obj))


def format_date(dt, format='%Y-%m-%d %H:%M:%S'):
    """
    Args:
        dt:
        format:
    """
    if dt is None:
        return ''
    local_zone = tz.tzlocal()
    # make sure this is UTC
    dt = dt.replace(tzinfo=tz.tzutc())
    local_time = dt.astimezone(local_zone)
    return local_time.strftime(format)


def format_timedelta(td):
    """
    Args:
        td:
    """
    if td is None:
        return ''
    hours, remainder = divmod(td.total_seconds(), 3600)
    minutes, seconds = divmod(remainder, 60)
    return '{:02}:{:02}:{:02}'.format(int(hours), int(minutes), int(seconds))


def parse_unknown_options(args):
    """
    Args:
        args:
    """
    warning_message = 'Please provide args with a proper ' \
                      'identifier as the key and the following structure: ' \
                      '--custom_argument="value"'

    assert all(a.startswith('--') for a in args), warning_message
    assert all(len(a.split('=')) == 2 for a in args), warning_message

    p_args = [a.lstrip('--').split('=') for a in args]

    assert all(k.isidentifier() for k, _ in p_args), warning_message

    r_args = {k: v for k, v in p_args}
    assert len(p_args) == len(r_args), 'Replicated arguments!'

    return r_args



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from abc import abstractmethod
from typing import Text, List

from tensorflow_metadata.proto.v0.schema_pb2 import Schema
from tensorflow_metadata.proto.v0.statistics_pb2 import \
    DatasetFeatureStatisticsList

from zenml.steps import BaseStep
from zenml.enums import StepTypes


class BaseSplitStep(BaseStep):
    """
    Base split class. Each custom data split should derive from this.
    In order to define a custom split, override the base split's partition_fn
    method.
    """

    STEP_TYPE = StepTypes.split.name

    def __init__(self,
                 statistics: DatasetFeatureStatisticsList = None,
                 schema: Schema = None,
                 **kwargs):
        """
        Base Split constructor.

        Args:
            statistics: Parsed statistics output of a preceding StatisticsGen.
            schema: Parsed schema output of a preceding SchemaGen.
        """
        super().__init__(**kwargs)
        self.statistics = statistics
        self.schema = schema

    @abstractmethod
    def partition_fn(self, element, n):
        """
        The partition function

        To be eligible to use in a Split Step, the partition_fn has to adhere
        to the following design contract:

        The signature is of the following type:

            >>> def partition_fn(element, n) -> int,

            where n is the number of splits and element is a data point,
            given as a tf.train.Example.

        Returns:
             signed integers i less than n, 0 ≤ i ≤ n - 1, indicating the
             split

        """
        pass

    @abstractmethod
    def get_split_names(self) -> List[Text]:
        """
        Returns the names of the splits associated with this split step.

        Returns:
            A list of strings, which are the split names.
        """
        pass

    def get_num_splits(self):
        """
        Returns the total number of splits.

        Returns:
            A positive integer, the number of splits.
        """
        return len(self.get_split_names())



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from abc import abstractmethod
from typing import Dict, List
from typing import Text

import tensorflow_transform as tft
from tensorflow_transform.tf_metadata import schema_utils
from tfx.utils import io_utils

from zenml.enums import StepTypes
from zenml.steps import BaseStep
from zenml.steps.trainer.utils import TRAIN_SPLITS, TEST_SPLITS, EVAL_SPLITS


class BaseTrainerStep(BaseStep):
    """
    Base step interface for all Trainer steps. All of your code concerning
    model training should leverage subclasses of this class.
    """

    STEP_TYPE = StepTypes.trainer.name

    def __init__(self,
                 input_patterns: Dict[Text, Text] = None,
                 output_patterns: Dict[Text, Text] = None,
                 serving_model_dir: Text = None,
                 transform_output: Text = None,
                 schema_path: Text = None,
                 split_mapping: Dict[Text, List[Text]] = None,
                 **kwargs):
        """
        Constructor for the BaseTrainerStep. All subclasses used for custom
        training of user machine learning models should implement the `run_fn`
        `model_fn` and `input_fn` methods used for control flow, model training
        and input data preparation, respectively.

        Args:
            serving_model_dir: Directory indicating where to save the trained
             model.
            transform_output: Output of a preceding transform component.
            train_files: String, file pattern of the location of TFRecords for
             model training. Intended for use in the input_fn.
            eval_files: String, file pattern of the location of TFRecords for
             model evaluation. Intended for use in the input_fn.
            log_dir: Logs output directory.
            schema: Schema file from a preceding SchemaGen.
        """
        # Inputs
        self.input_patterns = input_patterns
        self.schema = None
        self.tf_transform_output = None

        if schema_path is not None:
            schema = io_utils.SchemaReader().read(schema_path)
            self.schema = schema_utils.schema_as_feature_spec(
                schema).feature_spec

        # transform_output takes precendence over schema_path
        if transform_output is not None:
            self.tf_transform_output = tft.TFTransformOutput(transform_output)
            self.schema = self.tf_transform_output.transformed_feature_spec()

        # Parameters
        if split_mapping:
            assert len(split_mapping[TRAIN_SPLITS]) > 0, \
                'While defining your own mapping, you need to provide at ' \
                'least ' \
                'one training split.'
            assert len(split_mapping[EVAL_SPLITS]) > 0, \
                'While defining your own mapping, you need to provide at ' \
                'least ' \
                'one eval split.'

            if TEST_SPLITS not in split_mapping:
                split_mapping.update({TEST_SPLITS: []})
            assert len(split_mapping) == 3, \
                f'While providing a split_mapping please only use ' \
                f'{TRAIN_SPLITS}, {EVAL_SPLITS} and {TEST_SPLITS} as keys.'

            self.split_mapping = split_mapping
        else:
            self.split_mapping = {TRAIN_SPLITS: ['train'],
                                  EVAL_SPLITS: ['eval'],
                                  TEST_SPLITS: ['test']}

        # Outputs
        self.output_patterns = output_patterns
        self.serving_model_dir = serving_model_dir
        self.log_dir = None

        if self.serving_model_dir is not None:
            self.log_dir = os.path.join(
                os.path.dirname(self.serving_model_dir), 'logs')

        super(BaseTrainerStep, self).__init__(split_mapping=self.split_mapping,
                                              **kwargs)

    def input_fn(self):
        """
        Method for loading data from TFRecords saved to a location on
        disk. Override this method in subclasses to define your own custom
        data preparation flow.

        Returns:
            dataset: A tf.data.Dataset constructed from the input file
             pattern and transform.
        """
        pass

    def model_fn(self, *args, **kwargs):
        """
        Method defining the training flow of the model. Override this
        in subclasses to define your own custom training flow.

        Returns:
            model: A trained machine learning model.
        """
        pass

    def test_fn(self):
        """
        Optional method for defining a test flow of the model. The goal of
        this method is to give the user an interface to provide a testing
        function, where the results (if given in the right format with
        features, labels and predictions) will be ultimately saved to
        disk using an output artifact. Once defined, it allows the user to
        utilize the model agnostic evaluator in their training pipeline.
        """
        pass

    @abstractmethod
    def run_fn(self):
        """
        Method defining the control flow of the training process inside
        the TFX Trainer Component Executor. Override this method in subclasses
        to define your own custom training flow.
        """
        pass



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from pathlib import Path

import zenml

# Nicholas a way to get to the root
ZENML_ROOT = str(Path(zenml.__path__[0]).parent)
TEST_ROOT = os.path.join(ZENML_ROOT, "tests")

# TODO [LOW]: These require GOOGLE_APPLICATION_CREDENTIALS in the runner so
#  we take out for now.

# from zenml.datasources import BigQueryDatasource
# def test_bq_public_commit(repo):
#     ds = BigQueryDatasource(
#         name='bq_datasource_public',
#         query_project='bigquery-public-data',
#         query_dataset='samples',
#         query_table='shakespeare',
#         gcs_location='gs://zenml_tests/bq_datasource_test',
#         dest_project='maiot-zenml',
#         query_limit=100,
#     )
#     assert not ds.commits
#
#     _id = ds.commit()
#
#     # check that this commit made
#     assert _id == ds.get_latest_commit()
#
#
# def test_bq_private_commit(repo):
#     ds = BigQueryDatasource(
#         name='bq_datasource_private',
#         query_project='maiot-zenml',
#         query_dataset='zenml_tests',
#         query_table='shakespeare',
#         gcs_location='gs://zenml_tests/bq_datasource_test',
#         query_limit=100,
#     )
#     assert not ds.commits
#
#     _id = ds.commit()
#
#     # check that this commit made
#     assert _id == ds.get_latest_commit()



from step.trainer import MyPyTorchLightningTrainer
from zenml.datasources import CSVDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository
from zenml.steps.evaluator import AgnosticEvaluator
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.utils import naming_utils

# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = CSVDatasource(name='Pima Indians Diabetes',
                       path='gs://zenml_quickstart/diabetes.csv')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'Pima Indians Diabetes')
training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    StandardPreprocesser(
        features=['times_pregnant', 'pgc', 'dbp', 'tst', 'insulin', 'bmi',
                  'pedigree', 'age'],
        labels=['has_diabetes'],
        overwrite={'has_diabetes': {
            'transform': [{'method': 'no_transform', 'parameters': {}}]}}
    ))

# Add a trainer
training_pipeline.add_trainer(MyPyTorchLightningTrainer(epochs=10))

# Add an evaluator
label_name = naming_utils.transformed_label_name('has_diabetes')
training_pipeline.add_evaluator(
    AgnosticEvaluator(
        prediction_key='output',
        label_key=label_name,
        slices=[['has_diabetes']],
        metrics=['mean_squared_error']))

# Run the pipeline locally
training_pipeline.run()

# Evaluate
training_pipeline.evaluate()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the categorical domain split."""

from typing import Text, Dict, List, Any, Union

from zenml.steps.split import constants
from zenml.steps.split import BaseSplitStep
from zenml.steps.split.utils import get_categorical_value

CategoricalValue = Union[Text, int]


def lint_split_map(split_map: Dict[Text, List[CategoricalValue]]):
    """Small utility to lint the split_map"""
    if len(split_map) <= 1:
        raise AssertionError('Please specify more than 1 split name in the '
                             'split_map!')


class CategoricalDomainSplit(BaseSplitStep):
    """
    Categorical domain split. Use this to split data based on values in
    a single categorical column. A categorical column is defined here as a
    column with finitely many values of type `integer` or `string`.
    """

    def __init__(
            self,
            categorical_column: Text,
            split_map: Dict[Text, List[CategoricalValue]],
            unknown_category_policy: Text = constants.SKIP,
            statistics=None,
            schema=None,
    ):
        """
        Categorical domain split constructor.

        Use this class to split your data based on values in
        a single categorical column. A categorical column is defined here as a
        column with finitely many values of type `integer` or `string`.

        Example usage:

        # Split on a categorical attribute called "color".

        # red and blue datapoints go into the train set,
           green and yellow ones go into the eval set. Other colors,
           e.g. "purple", are discarded due to the "skip" flag.

        >>> split = CategoricalDomainSplit(
        ... categorical_column="color",
        ... split_map = {"train": ["red", "blue"],
        ...              "eval": ["green", "yellow"]},
        ... unknown_category_policy="skip")

        Supply the ``unknown_category_policy`` flag to set the unknown
        category handling policy. There are two main options:

        Setting ``unknown_category_policy`` to any key in the split map
        indicates that any missing categories should be put into that
        particular split. For example, supplying
        ``unknown_category_policy="train"`` indicates that all missing
        categories should go into the training dataset, while
        ``unknown_category_policy="eval"`` indicates that all missing
        categories should go into the evaluation dataset.

        Setting ``unknown_category_policy="skip"`` indicates that data points
        with unknown categorical values (i.e., values not present in any of the
        categorical value lists inside the split map) should be taken out of
        the data set.

        Args:
            statistics: Parsed statistics artifact from a preceding
             StatisticsGen.
            schema: Parsed schema artifact from a preceding SchemaGen.
            categorical_column: Name of the categorical column in the data on
             which to split.
            split_map: A dict { split_name: [categorical_values] } mapping
             categorical values to their respective splits.
            unknown_category_policy: String, indicates how to handle categories
             in the data that are not present in the split map.
        """
        self.categorical_column = categorical_column

        lint_split_map(split_map)
        self.split_map = split_map

        if unknown_category_policy in self.split_map:
            self.unknown_category_policy = unknown_category_policy
        else:
            self.unknown_category_policy = constants.SKIP

        super().__init__(statistics=statistics,
                         schema=schema,
                         categorical_column=categorical_column,
                         split_map=split_map,
                         unknown_category_policy=unknown_category_policy)

    def partition_fn(self,
                     element: Any,
                     num_partitions: int) -> int:

        """
        Function for a categorical split on data to be used in a beam.Partition.
        Args:
            element: Data point, given as a tf.train.Example.
            num_partitions: Number of splits, unused here.

        Returns:
            An integer n, where 0 ≤ n ≤ num_partitions - 1.

        """

        category_value = get_categorical_value(element,
                                               cat_col=self.categorical_column)

        # The following code produces a dict: { split_name: unique_integer }
        enumerated_splits = {name: i for i, name in enumerate(self.split_map.keys())}

        if self.unknown_category_policy not in enumerated_splits:
            enumerated_splits.update({self.unknown_category_policy: num_partitions - 1})

        for split_name, category_value_list in self.split_map.items():
            # if the value is in the list, then just return
            if category_value in category_value_list:
                # return the index of that split
                return enumerated_splits[split_name]

        # This is the default behavior for category_values that dont belong to
        #  any split in the split_map.
        return enumerated_splits[self.unknown_category_policy]

    def get_split_names(self) -> List[Text]:
        split_names = list(self.split_map.keys())
        if self.unknown_category_policy in self.split_map:
            return split_names
        else:
            return split_names + [constants.SKIP]



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of an Artifact Store"""

import hashlib
import os
from typing import Text

from zenml.repo import GlobalConfig
from zenml.utils import path_utils
from zenml.enums import ArtifactStoreTypes
from zenml.utils.print_utils import to_pretty_string, PrintStyles


class ArtifactStore:
    """Base class for all ZenML datasources.

    Every ZenML datasource should override this class.
    """

    def __init__(self, path: Text):
        self.path = path

        if path_utils.is_gcs_path(path):
            self.store_type = ArtifactStoreTypes.gcs.name
        else:
            self.store_type = ArtifactStoreTypes.local.name

        # unique_id based on path should be globally consistent
        self.unique_id = hashlib.md5(self.path.encode()).hexdigest()

    def __str__(self):
        return to_pretty_string(self.path)

    def __repr__(self):
        return to_pretty_string(self.path, style=PrintStyles.PPRINT)

    @staticmethod
    def get_component_name_from_uri(artifact_uri: Text):
        """
        Gets component name from artifact URI.

        Args:
            artifact_uri (str): URI to artifact.
        """
        return path_utils.get_grandparent(artifact_uri)

    def resolve_uri_locally(self, artifact_uri: Text, path: Text = None):
        """
        Takes a URI that points within the artifact store, downloads the
        URI locally, then returns local URI.

        Args:
            artifact_uri: uri to artifact.
            path: optional path to download to. If None, is inferred.
        """
        if not path_utils.is_remote(artifact_uri):
            # Its already local
            return artifact_uri

        if path is None:
            # Create a unique path in local machine
            path = os.path.join(
                GlobalConfig.get_config_dir(),
                self.unique_id,
                ArtifactStore.get_component_name_from_uri(artifact_uri),
                path_utils.get_parent(artifact_uri)  # unique ID from MLMD
            )

        # Create if not exists and download
        path_utils.create_dir_recursive_if_not_exists(path)
        path_utils.copy_dir(artifact_uri, path, overwrite=True)

        return path



import os

import numpy as np
import pandas as pd
import tensorflow as tf

from zenml.utils import path_utils

SPLIT_MAPPING = 'split_mapping'

TRAIN_SPLITS = 'train_splits'
EVAL_SPLITS = 'eval_splits'
TEST_SPLITS = 'test_splits'


def save_test_results(results, output_path):
    path_utils.create_dir_if_not_exists(output_path)
    output_path = os.path.join(output_path, 'test_results')
    with tf.io.TFRecordWriter(output_path) as writer:
        for example in to_serialized_examples(results):
            writer.write(example)

    return results


def to_serialized_examples(instance) -> tf.train.Example:
    example_list = []

    keys = list(instance.keys())
    num_samples = instance[keys[0]].shape[0]

    for i in range(num_samples):
        feature = {}
        for key in keys:
            value = instance[key][i]

            if value is None:
                raise Exception('The value can not possibly be None!')

            if len(value.shape) < 1:
                value = np.array([value])

            if pd.api.types.is_integer_dtype(value):
                feature[key] = tf.train.Feature(
                    int64_list=tf.train.Int64List(value=value))
            elif pd.api.types.is_float_dtype(value):
                feature[key] = tf.train.Feature(
                    float_list=tf.train.FloatList(value=value))
            else:
                vec_F = np.vectorize(tf.compat.as_bytes)
                feature[key] = tf.train.Feature(
                    bytes_list=tf.train.BytesList(value=vec_F(value)))
        example_list.append(tf.train.Example(
            features=tf.train.Features(feature=feature)
        ).SerializeToString())

    return example_list


def combine_batch_results(x):
    import torch

    result = {}
    for batch in x:
        for feature, values in batch.items():
            if isinstance(values, torch.Tensor):
                temp = torch.squeeze(values).detach().numpy()
                values = np.reshape(temp, values.size())
            elif isinstance(values, tf.Tensor):
                values = values.numpy()

            if feature not in result:
                result[feature] = values
            else:
                result[feature] = np.concatenate((result[feature],
                                                  values), axis=0)

    return result



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


from typing import Dict, Any, Text

from tfx import types
from tfx.dsl.components.base import executor_spec
from tfx.dsl.components.base.base_component import BaseComponent
from tfx.types import standard_artifacts
from tfx.types.component_spec import ComponentSpec, ExecutionParameter, \
    ChannelParameter

from zenml.components.sequencer import constants, executor
from zenml.standards.standard_keys import StepKeys


class SequencerComponentSpec(ComponentSpec):
    """Sequencer component spec."""
    PARAMETERS = {
        StepKeys.SOURCE: ExecutionParameter(type=Text),
        StepKeys.ARGS: ExecutionParameter(type=Dict[Text, Any]),
    }
    INPUTS = {
        constants.INPUT_EXAMPLES: ChannelParameter(
            type=standard_artifacts.Examples),
        constants.STATISTICS: ChannelParameter(
            type=standard_artifacts.ExampleStatistics),
        constants.SCHEMA: ChannelParameter(
            type=standard_artifacts.Schema)
    }
    OUTPUTS = {
        constants.OUTPUT_EXAMPLES: ChannelParameter(
            type=standard_artifacts.Examples)
    }


class Sequencer(BaseComponent):
    SPEC_CLASS = SequencerComponentSpec
    EXECUTOR_SPEC = executor_spec.ExecutorClassSpec(executor.Executor)

    def __init__(self,
                 source: Text,
                 source_args: Dict[Text, Any],
                 input_examples: types.Channel,
                 statistics: types.Channel = None,
                 schema: types.Channel = None):
        output_examples = types.Channel(type=standard_artifacts.Examples)

        spec = SequencerComponentSpec(
            source=source,
            args=source_args,
            input_examples=input_examples,
            output_examples=output_examples,
            statistics=statistics,
            schema=schema)

        super(Sequencer, self).__init__(spec=spec)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""
.. currentmodule:: ce_cli.cli
.. moduleauthor:: maiot GmbH <support@maiot.io>
"""

import click

from zenml.cli.utils import pass_config
from zenml.logger import set_root_verbosity


@click.group()
@pass_config
def cli(info):
    """maiot ZenML"""
    info.load()
    set_root_verbosity()


if __name__ == '__main__':
    cli()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.steps.trainer import BaseTrainerStep


class TFBaseTrainerStep(BaseTrainerStep):
    """
    Base class for all Tensorflow-based trainer steps. All tensorflow based
    trainings should use this as the base class. An example is available
    with tf_ff_trainer.FeedForwardTrainer.
    """

    @staticmethod
    def _get_ce_eval_tf_examples_fn(model, tf_transform_output):
        """
        Serving evaluation signature definition. Override this function to
        define a custom model endpoint used for evaluation at serving time.

        Args:
            model: Trained model, output of the model_fn.
            tf_transform_output: Output of the preceding Preprocessing
             component.

        Returns:
            eval_examples_fn: A @tf.function annotated callable that takes
             in a serialized tf.train.Example, parses it and serves the model
             output of this Example as evaluation. In Tensorflow Extended,
             such a mechanism is called a signature; for an example
             implementation of this signature, see a ZenML derived TrainerStep
             class.
        """
        pass

    @staticmethod
    def _get_serve_tf_examples_fn(model, tf_transform_output):
        """
        Defines a serving prediction signature. Override this function to
        create a custom model endpoint that can be used for prediction at
        serving time.

        Args:
            model: Trained model, output of the model_fn.
            tf_transform_output: Output of the preceding Preprocessing
             component.

        Returns:
            serve_examples_fn: A @tf.function annotated callable that takes
             in a serialized tf.train.Example, parses it and serves the model
             output of this Example as a prediction. In Tensorflow Extended,
             such a mechanism is called a signature; for an example
             implementation of this signature, see a ZenML derived TrainerStep
             class.
        """
        pass





#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

def forward_f():
    def apply(x):
        x = x.ffill()
        x = x.bfill()
        return x.fillna(infer_default_value(x))

    return apply


def backwards_f():
    def apply(x):
        x = x.bfill()
        x = x.ffill()
        return x.fillna(infer_default_value(x))

    return apply


def custom_f(custom_value):
    def apply(x):
        x = x.fillna(custom_value)
        return x

    return apply


def min_f():
    def apply(x):
        x = x.fillna(x.min())
        return x.fillna(infer_default_value(x))

    return apply


def max_f():
    def apply(x):
        x = x.fillna(x.max())
        return x.fillna(infer_default_value(x))

    return apply


def mean_f():
    def apply(x):
        x = x.fillna(x.mean())
        return x.fillna(infer_default_value(x))

    return apply


def infer_default_value(x):
    if x.dtype == float:
        return 0.0
    else:
        return 'None'



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

import click

from zenml.cli.cli import cli
from zenml.utils import path_utils


def get_examples_dir():
    cli_dir = os.path.dirname(__file__)
    package_dir = os.path.dirname(os.path.dirname(cli_dir))
    examples_dir = os.path.join(package_dir, 'examples')
    return examples_dir


def get_all_examples():
    examples = []
    for name in sorted(os.listdir(get_examples_dir())):
        # Skip hidden files (like .gitignore)
        if not name.startswith('.') and not name.startswith(
                '__') and not name.startswith('README'):
            examples.append(name)
    return examples


def get_example_readme(example_path):
    with open(os.path.join(example_path, 'README.md')) as readme:
        readme_content = readme.read()
    return readme_content


@cli.group(help="Access all ZenML examples.")
def example():
    """Examples group"""
    pass


@example.command(help="List the available examples.")
def list():
    click.echo('Listing examples: \n')
    for name in get_all_examples():
        click.echo(f'{name}')
    click.echo('\nTo pull the examples, type: ')
    click.echo('zenml example pull EXAMPLE_NAME')


@example.command(help='Find out more about an example.')
@click.argument('example_name')
def info(example_name):
    example_dir = os.path.join(get_examples_dir(), example_name)
    readme_content = get_example_readme(example_dir)
    click.echo(readme_content)


@example.command(help="Pull examples straight " \
                       "into your current working directory.")
@click.argument('example_name', required=False, default=None)
def pull(example_name):
    examples_dir = get_examples_dir()
    if not example_name:
        examples = get_all_examples()
    else:
        examples = [example_name]

    # Create destination dir.
    dst = os.path.join(os.getcwd(), 'zenml_examples')
    path_utils.create_dir_if_not_exists(dst)

    # Pull specified examples.
    for example in examples:
        dst_dir = os.path.join(dst, example)
        # Check if example has already been pulled before.
        if path_utils.file_exists(dst_dir):
            if click.confirm(f'Example {example} is already pulled. '
                             f'Do you wish to overwrite the directory?'):
                path_utils.rm_dir(dst_dir)
            else:
                continue
        click.echo(f'Pulling example {example}')
        src_dir = os.path.join(examples_dir, example)
        path_utils.copy_dir(src_dir, dst_dir)

        click.echo(f'Example pulled in directory: {dst_dir}')

    click.echo()
    click.echo('Please read the README.md file in the respective example '
               'directory to find out more about the example')



from step.trainer import TorchRegressionTrainer
from zenml.datasources import CSVDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository
from zenml.steps.evaluator import AgnosticEvaluator
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.utils import naming_utils

training_pipeline = TrainingPipeline()

try:
    ds = CSVDatasource(name='boston_data', path='./data/boston_data.csv')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'boston_data')
training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    StandardPreprocesser(
        features=['crim',
                  'zn',
                  'indus',
                  'chas',
                  'nox',
                  'rm',
                  'age',
                  'dis',
                  'rad',
                  'tax',
                  'ptratio',
                  'black',
                  'lstat'],
        labels=['medv'],
        overwrite={'medv': {
            'transform': [{'method': 'no_transform', 'parameters': {}}]}}
    ))

# Add a trainer

training_pipeline.add_trainer(TorchRegressionTrainer(
    loss='mse',
    optimizer="adam",
    last_activation='relu',
    metrics=['accuracy'],
    epochs=100))

# Add an evaluator
label_name = naming_utils.transformed_label_name('medv')
training_pipeline.add_evaluator(
    AgnosticEvaluator(
        prediction_key='output',
        label_key=label_name,
        slices=[['medv']],
        metrics=['mean_squared_error']))

# Run the pipeline locally
training_pipeline.run()
training_pipeline.evaluate()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import logging
import os
import sys
from logging.handlers import TimedRotatingFileHandler

from absl import logging as absl_logging

from zenml.constants import ZENML_LOGGING_VERBOSITY, APP_NAME, \
    ABSL_LOGGING_VERBOSITY

# Mute tensorflow cuda warnings
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

LOGGING_LEVELS = {
    0: logging.NOTSET,
    1: logging.ERROR,
    2: logging.WARN,
    3: logging.INFO,
    4: logging.DEBUG,
}

FORMATTING_STRING = "%(asctime)s — %(name)s — %(levelname)s — %(message)s"
FORMATTER = logging.Formatter(FORMATTING_STRING)
LOG_FILE = f'{APP_NAME}_logs.log'


def resolve_logging_level():
    if ZENML_LOGGING_VERBOSITY > 0:
        level = LOGGING_LEVELS[ZENML_LOGGING_VERBOSITY] \
            if ZENML_LOGGING_VERBOSITY in LOGGING_LEVELS else logging.DEBUG
    else:
        level = logging.NOTSET
    return level


def set_root_verbosity():
    level = resolve_logging_level()
    if level > logging.NOTSET:
        logging.basicConfig(level=level)
        get_logger(__name__).debug(f'Logging set to level: '
                                   f'{logging.getLevelName(level)}')
    else:
        logging.disable(sys.maxsize)
        logging.getLogger().disabled = True
        get_logger(__name__).debug('Logging NOTSET')


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    file_handler = TimedRotatingFileHandler(LOG_FILE, when='midnight')
    file_handler.setFormatter(FORMATTER)
    return file_handler


def get_logger(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(resolve_logging_level())
    logger.addHandler(get_console_handler())

    # TODO: [LOW] Add a file handler for persistent handling
    # logger.addHandler(get_file_handler())
    # with this pattern, it's rarely necessary to propagate the error up to
    # parent
    logger.propagate = False
    return logger


def init_logging():
    logging.basicConfig(format=FORMATTING_STRING)
    set_root_verbosity()

    # set absl logging
    absl_logging.set_verbosity(ABSL_LOGGING_VERBOSITY)



import os

from predictor.tf import TensorFlowPredictor
from zenml.datasources import CSVDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.pipelines import TrainingPipeline
from zenml.repo import ArtifactStore
from zenml.steps.deployer import CortexDeployer
from zenml.steps.evaluator import TFMAEvaluator
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.utils.naming_utils import transformed_label_name

GCP_BUCKET = os.getenv('GCP_BUCKET')
assert GCP_BUCKET
CORTEX_ENV = os.getenv('CORTEX_ENV', 'env')
CORTEX_MODEL_NAME = os.getenv('CORTEX_MODEL_NAME', 'zenml-classifier')

# For this example, the ArtifactStore must be a GCP bucket, as the
# CortexDeployer step is using the GCP env.

from zenml.repo.repo import Repository

# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = CSVDatasource(name='Pima Indians Diabetes',
                       path='gs://zenml_quickstart/diabetes.csv')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'Pima Indians Diabetes')
training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    StandardPreprocesser(
        features=['times_pregnant', 'pgc', 'dbp', 'tst', 'insulin', 'bmi',
                  'pedigree', 'age'],
        labels=['has_diabetes'],
        overwrite={'has_diabetes': {
            'transform': [{'method': 'no_transform', 'parameters': {}}]}}
    ))

# Add a trainer
training_pipeline.add_trainer(TFFeedForwardTrainer(
    loss='binary_crossentropy',
    last_activation='sigmoid',
    output_units=1,
    metrics=['accuracy'],
    epochs=15))

# Add an evaluator
training_pipeline.add_evaluator(
    TFMAEvaluator(slices=[['has_diabetes']],
                  metrics={transformed_label_name('has_diabetes'): [
                      'binary_crossentropy',
                      'binary_accuracy']}))
# Add cortex deployer
api_config = {
    "name": CORTEX_MODEL_NAME,
    "kind": "RealtimeAPI",
    "predictor": {
        "type": "tensorflow",
        # Set signature key of the model as we are using Tensorflow Trainer
        "models": {"signature_key": "serving_default"}}
}
training_pipeline.add_deployment(
    CortexDeployer(
        env=CORTEX_ENV,
        api_config=api_config,
        predictor=TensorFlowPredictor,
    )
)

# Define the artifact store
artifact_store = ArtifactStore(
    os.path.join(GCP_BUCKET, 'cortex/artifact_store'))

# Run the pipeline
training_pipeline.run(artifact_store=artifact_store)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Orchestrator for simple AWS VM backend"""

import os
import time
from typing import Text, Dict, Any

from zenml.backends.orchestrator.aws import utils
from zenml.backends.orchestrator import OrchestratorBaseBackend
from zenml.repo import Repository
from zenml.standards import standard_keys as keys
from zenml.utils import path_utils
from zenml.constants import ZENML_BASE_IMAGE_NAME
from zenml.logger import get_logger

logger = get_logger(__name__)

EXTRACTED_TAR_DIR_NAME = 'zenml_working'
STAGING_AREA = 'staging'
TAR_PATH_ARG = 'tar_path'


class OrchestratorAWSBackend(OrchestratorBaseBackend):
    """
    Orchestrates pipeline on a AWS EC2 instance
    """

    def __init__(self,
                 iam_role: Text,
                 instance_type: Text = 't2.micro',
                 instance_image: Text = 'ami-02e9f4e447e4cda79',
                 zenml_image: Text = None,
                 region: Text = None,
                 key_name: Text = None,
                 security_group: Text = None,
                 min_count: int = 1,
                 max_count: int = 1,
                 **kwargs):
        """
        Base class for the orchestrator backend on AWS

        :param iam_role: the name of the role created in AWS IAM
        :param instance_type: the type of the EC2 instance, defaults to
        t2.micro
        :param instance_image: the image for the EC2 instance, defaults to the
        public image: Deep Learning AMI (Amazon Linux 2) Version 39.0
        :param zenml_image: refers to the image with ZenML
        :param region: the name of the region that AWS is working on
        :param key_name: the name of the key to be used whilst creating the
        instance on EC2
        :param security_group: the name of a selected security group
        :param min_count: the minimum number of instances, defaults to 1
        :param max_count: the maximum number of instances, defaults to 1
        """

        self.session = utils.setup_session()
        self.region = utils.setup_region(region)

        self.ec2_client = self.session.client('ec2')
        self.ec2_resource = self.session.resource('ec2')

        self.instance_type = instance_type
        self.instance_image = instance_image
        self.zenml_image = zenml_image
        self.key_name = key_name
        self.min_count = min_count
        self.max_count = max_count

        if security_group is not None:
            self.security_group = [security_group]
        else:
            self.security_group = security_group

        self.iam_role = {'Name': iam_role}

        if zenml_image is None:
            self.zenml_image = ZENML_BASE_IMAGE_NAME
        else:
            self.zenml_image = zenml_image

        super(OrchestratorBaseBackend, self).__init__(
            instance_type=self.instance_type,
            instance_image=self.instance_image,
            zenml_image=self.zenml_image,
            region=self.region,
            key_name=self.key_name,
            min_count=self.min_count,
            max_count=self.max_count,
            security_group=self.security_group,
            iam_role=self.iam_role,
            **kwargs,
        )

    @staticmethod
    def make_unique_name(name):
        return f'{name}-{time.asctime()}'

    def launch_instance(self, config):
        startup = utils.get_startup_script(config,
                                           self.region,
                                           self.zenml_image)

        args = {'ImageId': self.instance_image,
                'InstanceType': self.instance_type,
                'IamInstanceProfile': self.iam_role,
                'MaxCount': self.max_count,
                'MinCount': self.min_count,
                'UserData': startup}

        if self.security_group:
            args['SecurityGroups'] = self.security_group

        if self.key_name:
            args['KeyName'] = self.key_name

        return self.ec2_resource.create_instances(**args)

    def run(self, config: [Dict, Any]):
        # Extract the paths to create the tar
        logger.info('Orchestrating pipeline on AWS..')

        repo: Repository = Repository.get_instance()
        repo_path = repo.path
        config_dir = repo.zenml_config.config_dir
        tar_file_name = \
            f'{EXTRACTED_TAR_DIR_NAME}_{str(int(time.time()))}.tar.gz'
        path_to_tar = os.path.join(config_dir, tar_file_name)

        # Create tarfile but exclude .zenml folder if exists
        path_utils.create_tarfile(repo_path, path_to_tar)
        logger.info(f'Created tar of current repository at: {path_to_tar}')

        # Upload tar to artifact store
        store_path = config[keys.GlobalKeys.ARTIFACT_STORE]
        store_staging_area = os.path.join(store_path, STAGING_AREA)
        store_path_to_tar = os.path.join(store_staging_area, tar_file_name)
        path_utils.copy(path_to_tar, store_path_to_tar)
        logger.info(f'Copied tar to artifact store at: {store_path_to_tar}')

        # Remove tar
        path_utils.rm_dir(path_to_tar)
        logger.info(f'Removed tar at: {path_to_tar}')

        # Append path of tar in config orchestrator utils
        config[keys.GlobalKeys.BACKEND][keys.BackendKeys.ARGS][
            TAR_PATH_ARG] = store_path_to_tar

        # Launch the instance
        self.launch_instance(config)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""BigQuery Datasource definition"""

from typing import Optional, Callable
from typing import Text, Dict

from apache_beam.io.gcp import bigquery as beam_bigquery

from zenml.datasources import BaseDatasource
from zenml.utils.beam_utils import WriteToTFRecord


class BigQueryDatasource(BaseDatasource):
    """ZenML BigQuery datasource definition.

    Use this for BigQuery training pipelines.
    """

    def __init__(
            self,
            name: Text,
            query_project: Text,
            query_dataset: Text,
            query_table: Text,
            gcs_location: Text,
            query_limit: Optional[int] = None,
            dest_project: Text = None,
            schema: Dict = None,
            **kwargs):
        """
        Initialize BigQuery source. This creates a datasource that
        essentially reflects the following query using Apache Beam.

        `SELECT * FROM query_project.query_dataset.query_table`

        A Google Cloud Storage location needs to be provided to make this work.
        The GCS location is used to write temporary dumps of the query as the
        beam pipeline executes. The location must exist within a GCP project
        specified through dest_project.

        Args:
            name: name of datasource. Must be globally unique in the repo.
            query_project: name of gcp project.
            query_dataset: name of dataset.
            query_table: name of table in dataset.
            query_limit: how many rows, from the top, to be queried.
            gcs_location: google cloud storage (bucket) location to store temp.
            dest_project: name of destination project. If None is specified,
            then dest_project is set to the same as query_project.
            schema (str): optional schema for data to conform to.
        """
        # Check whether gcs_location is a valid one
        if not gcs_location.startswith('gs://'):
            Exception(f'{gcs_location} is not a valid GCS path. It must start '
                      f'with a gs://')

        self.query_project = query_project
        self.query_dataset = query_dataset
        self.query_table = query_table
        self.query_limit = query_limit
        self.gcs_location = gcs_location
        self.schema = schema

        # If dest project not given, we use the same as query project
        self.dest_project = dest_project if dest_project else query_project

        super().__init__(
            name,
            query_project=self.query_project,
            query_dataset=self.query_dataset,
            query_table=self.query_table,
            gcs_location=self.gcs_location,
            query_limit=self.query_limit,
            dest_project=self.dest_project,
            schema=self.schema,
            **kwargs
        )

    def process(self, output_path: Text, make_beam_pipeline: Callable = None):
        query = f'SELECT * FROM `{self.query_project}.{self.query_dataset}.' \
                f'{self.query_table}`'

        if self.query_limit is not None:
            query += f'\nLIMIT {self.query_limit}'

        with make_beam_pipeline() as p:
            (p
             | 'ReadFromBigQuery' >> beam_bigquery.ReadFromBigQuery(
                        project=self.dest_project,
                        gcs_location=self.gcs_location,
                        query=query,
                        use_standard_sql=True)
             | WriteToTFRecord(self.schema, output_path)
             )



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import List, Text

import tensorflow_model_analysis as tfma
from google.protobuf.wrappers_pb2 import BoolValue

from zenml.steps.evaluator.base_evaluator import BaseEvaluatorStep


def to_camel_case(s):
    """
    Converts a given name to camel case and removes any '_'
    """
    return ''.join(list(map(lambda x: x.capitalize(), s.split('_'))))


class AgnosticEvaluator(BaseEvaluatorStep):
    """
    AgnosticEvaluator step designed to work regardless of the type of the
    trained model

    Through this step, it is possible to use the model_agnostic capabilities of
    TFMA, however this requires the test_results (tf.train.Example) to be saved
    beforehand in a flat format in the trainer step. Once the results are
    saved, you can use the label_key and the prediction_key to specify the
    selected features and compute the desired metrics.

    Note: As of now, it only supports a single label-prediction key pair, we
    are working hard to bring you support for multi-output models as well.
    """

    def __init__(self,
                 label_key: Text,
                 prediction_key: Text,
                 slices: List[List[Text]] = None,
                 metrics: List[Text] = None,
                 **kwargs):
        """
        Init for the AgnosticEvaluator

        :param label_key: string specifying the name of the label in the
        flattened datapoint
        :param prediction_key: string specifying the name of the output in the
        flattened datapoint
        :param slices: a list of lists, each element in the inner list include
        a set of features which will be used for slicing on the results
        :param metrics: list of metrics to be computed
        :param splits: the list of splits to apply the evaluation on
        """
        super(AgnosticEvaluator, self).__init__(label_key=label_key,
                                                prediction_key=prediction_key,
                                                slices=slices,
                                                metrics=metrics,
                                                **kwargs)

        self.slices = slices or list()
        self.metrics = metrics or list()

        self.prediction_key = prediction_key
        self.label_key = label_key

    def build_config(self):
        # SLICING SPEC
        slicing_specs = [tfma.SlicingSpec()]
        if self.slices:
            slicing_specs.extend([tfma.SlicingSpec(feature_keys=e)
                                  for e in self.slices])

        # MODEL SPEC
        model_specs = [tfma.ModelSpec(label_key=self.label_key,
                                      prediction_key=self.prediction_key)]

        # METRIC SPEC
        baseline = [tfma.MetricConfig(class_name='ExampleCount')]
        for key in self.metrics:
            baseline.append(tfma.MetricConfig(class_name=to_camel_case(key)))

        metrics_specs = [tfma.MetricsSpec(metrics=baseline)]

        return tfma.EvalConfig(
            model_specs=model_specs,
            slicing_specs=slicing_specs,
            metrics_specs=metrics_specs,
            options=tfma.Options(
                include_default_metrics=BoolValue(value=False)))



from zenml.standards.standard_keys import MethodKeys


def parse_methods(input_dict, process, methods):
    """
    Args:
        input_dict:
        process:
        methods:
    """
    result = {}

    # Obtain the set of features
    f_set = set(input_dict.keys())
    f_list = list(sorted(f_set))

    for feature in f_list:
        assert isinstance(input_dict[feature], dict), \
            'Please specify a dict for every feature (empty dict if default)'

        # Check if the process for the given feature has been modified
        if process in input_dict[feature].keys():
            result[feature] = []
            for m in input_dict[feature][process]:
                MethodKeys.key_check(m)
                method_name = m[MethodKeys.METHOD]
                parameters = m[MethodKeys.PARAMETERS]
                # Check if the selected method exists and whether the
                # right parameters are given
                methods.check_name_and_params(method_name, parameters)
                result[feature].append(m)

    return result


class MethodDescriptions:
    MODES = {}

    @classmethod
    def check_name_and_params(cls, method_name, method_params):
        """
        Args:
            method_name:
            method_params:
        """
        assert method_name in cls.MODES.keys(), \
            'Choose one of the defined methods: {}'.format(cls.MODES.keys())

        assert all(
            p in method_params.keys() for p in cls.MODES[method_name][1]), \
            'All the required params {} of the {} needs to be defined' \
                .format(cls.MODES[method_name][1], method_name)

    @classmethod
    def get_method(cls, method_name):
        """
        Args:
            method_name:
        """
        return cls.MODES[method_name][0]


DEFAULT_DICT = {
    'boolean': {
        'filling': [{'method': 'max', 'parameters': {}}],
        'resampling': [{'method': 'threshold',
                        'parameters': {'c_value': 0,
                                       'cond': 'greater',
                                       'set_value': 1,
                                       'threshold': 0}}],
        'transform': [{'method': 'no_transform', 'parameters': {}}]},
    'float': {
        'filling': [{'method': 'max', 'parameters': {}}],
        'resampling': [{'method': 'mean', 'parameters': {}}],
        'transform': [{'method': 'scale_to_z_score', 'parameters': {}}]},
    'integer': {
        'filling': [{'method': 'max', 'parameters': {}}],
        'resampling': [{'method': 'mean', 'parameters': {}}],
        'transform': [{'method': 'scale_to_z_score', 'parameters': {}}]},
    'string': {
        'filling': [{'method': 'custom', 'parameters': {'custom_value': ''}}],
        'resampling': [{'method': 'mode', 'parameters': {}}],
        'transform': [{'method': 'compute_and_apply_vocabulary',
                       'parameters': {}}]
    }}



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import List, Text

import tensorflow as tf
from transformers import TFDistilBertForSequenceClassification

from zenml.steps.trainer import TFBaseTrainerStep
from zenml.steps.trainer import utils


class UrduTrainer(TFBaseTrainerStep):
    def __init__(self,
                 model_name: Text,
                 batch_size: int = 64,
                 epochs: int = 25,
                 learning_rate: float = 1e-4,
                 **kwargs
                 ):
        super(UrduTrainer, self).__init__(model_name=model_name,
                                          batch_size=batch_size,
                                          epochs=epochs,
                                          learning_rate=learning_rate,
                                          **kwargs)

        # ML variables
        self.batch_size = batch_size
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.model_name = model_name

    def get_run_fn(self):
        return self.run_fn

    def run_fn(self):
        train_split_patterns = [self.input_patterns[split] for split in
                                self.split_mapping[utils.TRAIN_SPLITS]]
        train_dataset = self.input_fn(train_split_patterns)

        eval_split_patterns = [self.input_patterns[split] for split in
                               self.split_mapping[utils.EVAL_SPLITS]]
        eval_dataset = self.input_fn(eval_split_patterns)

        model = self.model_fn(train_dataset=train_dataset,
                              eval_dataset=eval_dataset)

        # Saving without signatures means no eval or deployer possible yet
        model.save_pretrained(self.serving_model_dir, saved_model=True)

    def model_fn(self,
                 train_dataset: tf.data.Dataset,
                 eval_dataset: tf.data.Dataset):
        id2label = {0: "FAKE_NEWS", 1: "REAL_NEWS"}

        model = TFDistilBertForSequenceClassification.from_pretrained(
            self.model_name, id2label=id2label)

        optimizer = tf.keras.optimizers.Adam(learning_rate=self.learning_rate)

        model.compile(optimizer=optimizer,
                      loss=model.compute_loss,
                      metrics=["accuracy"])

        model.fit(train_dataset,
                  epochs=self.epochs,
                  callbacks=[
                      tf.keras.callbacks.TensorBoard(
                          log_dir=os.path.join(self.log_dir, 'train'))])

        return model

    def input_fn(self,
                 file_pattern: List[Text]):
        """
        Feedforward input_fn for loading data from TFRecords saved to a
        location on disk.

        Args:
            file_pattern: File pattern matching saved TFRecords on disk.
            feature_spec: Feature spec from the inferred schema.

        Returns:
            dataset: tf.data.Dataset created out of the input files.
        """

        # grab BERT features plus the label
        bert_features = ["input_ids", "attention_mask"] + ["label"]

        feature_spec = {x: self.schema[x]
                        for x in self.schema
                        if x in bert_features}

        dataset = tf.data.experimental.make_batched_features_dataset(
            file_pattern=file_pattern,
            batch_size=self.batch_size,
            features=feature_spec,
            label_key="label",
            reader=self._gzip_reader_fn,
            num_epochs=1)

        return dataset

    @staticmethod
    def _gzip_reader_fn(filenames):
        """
        Small utility returning a record reader that can read gzipped files.

        Args:
            filenames: Names of the compressed TFRecord data files.
        """
        return tf.data.TFRecordDataset(filenames, compression_type='GZIP')

    def load_model(self, path_to_model: Text):
        return TFDistilBertForSequenceClassification.from_pretrained(
            path_to_model)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""JSON Datasource definition"""

from pathlib import Path
from typing import Text, Callable, Dict

from zenml.datasources import BaseDatasource
from zenml.utils import path_utils


class TFRecordsDatasource(BaseDatasource):
    """ZenML JSON datasource definition."""

    def __init__(
            self,
            name: Text,
            path: Text,
            schema: Dict = None,
            **kwargs):
        """
        Initialize TFRecords datasource. Data in the TFRecords should be stored
        in TFExamples that represent a list of dicts, where each dict
        represents one datapoint (including label, if required).

        E.g. The data can look like:
            [{
                'age': 20,
                'bmi': 33.3,
                'dbp': 72,
                'has_diabetes': 1,
                'insulin': 0,
                'pedigree': 0.627,
                'pgc': 168,
                'times_pregnant': 1,
                'tst': 35
            }]

        If TFRecords are stored in a different format, please override the
        `process` function of this class and write custom logic to convert
        to a format that looks like the above.

        Args:
            name: Name of datasource.
            path: path to directory of TFRecords files.
            schema (str): optional schema for data to conform to.
        """
        self.path = path
        self.schema = schema
        super().__init__(name, path=path, schema=schema, **kwargs)

    def process(self, output_path: Text, make_beam_pipeline: Callable = None):
        p = Path(self.path).resolve()
        path_utils.copy_dir(str(p), output_path)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


import os
from typing import List, Text, Dict

import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils.data as data
from torch.utils.tensorboard import SummaryWriter

from zenml.steps.trainer import TorchBaseTrainerStep
from zenml.steps.trainer import utils
from zenml.steps.trainer.pytorch_trainers import utils as torch_utils
from zenml.utils import path_utils


class BinaryClassifier(nn.Module):
    def __init__(self):
        super(BinaryClassifier, self).__init__()
        self.layer_1 = nn.Linear(8, 64)
        self.layer_2 = nn.Linear(64, 64)
        self.layer_out = nn.Linear(64, 1)

        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=0.1)
        self.batchnorm1 = nn.BatchNorm1d(64)
        self.batchnorm2 = nn.BatchNorm1d(64)

    def forward(self, inputs):
        x = self.relu(self.layer_1(inputs))
        x = self.batchnorm1(x)
        x = self.relu(self.layer_2(x))
        x = self.batchnorm2(x)
        x = self.dropout(x)
        x = self.layer_out(x)

        return x


def binary_acc(y_pred, y_test):
    y_pred_tag = torch.round(torch.sigmoid(y_pred))
    correct_results_sum = (y_pred_tag == y_test).sum().float()
    acc = correct_results_sum / y_test.shape[0]
    acc = torch.round(acc * 100)

    return acc


class FeedForwardTrainer(TorchBaseTrainerStep):
    def __init__(self,
                 batch_size: int = 32,
                 lr: float = 0.0001,
                 epochs: int = 10,
                 dropout_chance: int = 0.2,
                 loss: str = 'mse',
                 metrics: List[str] = None,
                 hidden_layers: List[int] = None,
                 hidden_activation: str = 'relu',
                 last_activation: str = 'sigmoid',
                 input_units: int = 8,
                 output_units: int = 1,
                 **kwargs):
        self.batch_size = batch_size
        self.lr = lr
        self.epochs = epochs
        self.dropout_chance = dropout_chance
        self.loss = loss
        self.metrics = metrics or []
        self.hidden_layers = hidden_layers or [64, 32, 16]
        self.hidden_activation = hidden_activation
        self.last_activation = last_activation
        self.input_units = input_units
        self.output_units = output_units

        super(FeedForwardTrainer, self).__init__(
            batch_size=self.batch_size,
            lr=self.lr,
            epochs=self.epochs,
            dropout_chance=self.dropout_chance,
            loss=self.loss,
            metrics=self.metrics,
            hidden_layers=self.hidden_layers,
            hidden_activation=self.hidden_activation,
            last_activation=self.last_activation,
            input_units=self.input_units,
            output_units=self.output_units,
            **kwargs)

    def input_fn(self,
                 file_patterns: List[Text]):

        dataset = torch_utils.TFRecordTorchDataset(file_patterns,
                                                   self.schema)
        loader = torch.utils.data.DataLoader(dataset,
                                             batch_size=self.batch_size,
                                             drop_last=True)
        return loader

    def model_fn(self, train_dataset, eval_dataset):
        return BinaryClassifier()

    def test_fn(self, model, dataset):
        # Activate the evaluation mode
        model.eval()

        batch_list = []
        for x, y, raw in dataset:
            # start with an empty batch
            batch = {}

            # add the raw features with the transformed features and labels
            batch.update(x)
            batch.update(y)
            batch.update(raw)

            # finally, add the output of the model
            x_batch = torch.cat([v for v in x.values()], dim=-1)
            p = model(x_batch)

            if isinstance(p, torch.Tensor):
                batch.update({'output': p})
            elif isinstance(p, dict):
                batch.update(p)
            elif isinstance(p, list):
                batch.update(
                    {'output_{}'.format(i): v for i, v in enumerate(p)})
            else:
                raise TypeError('Unknown output format!')

            batch_list.append(batch)

        combined_batch = utils.combine_batch_results(batch_list)

        return combined_batch

    def run_fn(self):
        train_split_patterns = [self.input_patterns[split] for split in
                                self.split_mapping[utils.TRAIN_SPLITS]]
        train_dataset = self.input_fn(train_split_patterns)

        eval_split_patterns = [self.input_patterns[split] for split in
                               self.split_mapping[utils.EVAL_SPLITS]]
        eval_dataset = self.input_fn(eval_split_patterns)

        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        model = self.model_fn(train_dataset, eval_dataset)

        model.to(device)
        criterion = nn.BCEWithLogitsLoss()
        optimizer = optim.Adam(model.parameters(), lr=0.001)

        writer = SummaryWriter(self.log_dir)

        model.train()

        total_count = 0

        for e in range(1, self.epochs + 1):
            epoch_loss = 0
            epoch_acc = 0
            step_count = 0
            for x, y, _ in train_dataset:
                step_count += 1
                total_count += 1

                x_batch = torch.cat([v.to(device) for v in x.values()], dim=-1)
                y_batch = torch.cat([v.to(device) for v in y.values()], dim=-1)
                optimizer.zero_grad()

                y_pred = model(x_batch)

                loss = criterion(y_pred, y_batch)
                acc = binary_acc(y_pred, y_batch)

                loss.backward()
                optimizer.step()

                epoch_loss += loss.item()
                epoch_acc += acc.item()

                if e == 1 and step_count == 1:
                    writer.add_graph(model, x_batch)

                writer.add_scalar('training_loss', loss, total_count)
                writer.add_scalar('training_accuracy', acc, total_count)

            print(f'Epoch {e + 0:03}: | Loss: '
                  f'{epoch_loss / step_count:.5f} | Acc: '
                  f'{epoch_acc / step_count:.3f}')

        # test
        for split in self.split_mapping[utils.TEST_SPLITS]:
            assert split in self.input_patterns, \
                f'There are currently no inputs for the split "{split}" ' \
                f'which is currently used in the {utils.TEST_SPLITS} of the ' \
                f'split mapping.'
            pattern = self.input_patterns[split]
            test_dataset = self.input_fn([pattern])
            test_results = self.test_fn(model, test_dataset)
            utils.save_test_results(test_results, self.output_patterns[split])

        path_utils.create_dir_if_not_exists(self.serving_model_dir)
        if path_utils.is_remote(self.serving_model_dir):
            temp_model_dir = '__temp_model_dir__'
            temp_path = os.path.join(os.getcwd(), temp_model_dir)
            if path_utils.is_dir(temp_path):
                raise PermissionError('{} is used as a temp path but it '
                                      'already exists. Please remove it to '
                                      'continue.')
            torch.save(model, temp_path)
            path_utils.copy_dir(temp_path, self.serving_model_dir)
            path_utils.rm_dir(temp_path)
        else:
            torch.save(model, os.path.join(self.serving_model_dir, 'model.pt'))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import datetime
import os
import re
import shutil
from distutils.core import run_setup
from tfx.utils import io_utils
from zenml.logger import get_logger

logger = get_logger(__name__)


def prepare_sdist():
    """
    Refer to the README.md in the docs folder
    """
    dist_path = os.path.join(os.getcwd(), 'dist')

    if os.path.exists(dist_path) and os.path.isdir(dist_path):
        logger.info('Removing {}'.format(dist_path))
        shutil.rmtree(dist_path)
    else:
        logger.warning('There is no dist folder.')

    # FOR THE BOPA
    run_setup('setup.py', script_args=['sdist'])
    req_path = os.path.join(dist_path, 'requirements.txt')
    io_utils.write_string_file(req_path, '')

    # FOR THE BEPA
    run_setup('setup.py', script_args=['sdist'])
    req_path = os.path.join(dist_path, 'requirements.txt')
    io_utils.write_string_file(req_path, '')


def generate_unique_name(base_name):
    """
    Args:
        base_name:
    """
    identifier = os.getenv('CI_COMMIT_SHORT_SHA', os.getenv('USER', 'local'))
    return re.sub(
        r'[^0-9a-zA-Z-]+',
        '-',
        '{pipeline_name}-{identifier}-{ts}'.format(
            pipeline_name=base_name,
            identifier=identifier,
            ts=int(datetime.datetime.timestamp(datetime.datetime.now()))
        ).lower()
    )


def sanitize_name_for_ai_platform(name: str):
    """
    Args:
        name (str):
    """
    return 'ce_' + name.replace('-', '_').lower()


def parse_yaml_beam_args(pipeline_args):
    """Converts yaml beam args to list of args TFX accepts

    Args:
        pipeline_args: dict specified in the config.yml

    Returns:
        list of strings, where each string is a beam argument
    """
    return ['--{}={}'.format(key, value) for key, value in
            pipeline_args.items()]



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import random

import pytest

from zenml.metadata import ZenMLMetadataStore
from zenml.standards.standard_keys import MLMetadataKeys
from zenml.enums import GDPComponent
from zenml.enums import PipelineStatusTypes

# we expect all queries to fail since the metadata store
# cannot be instantiated
expected_query_error = AssertionError


def test_metadata_init():
    mds1 = ZenMLMetadataStore()

    with pytest.raises(ValueError):
        _ = mds1.store


def test_to_config():
    mds1 = ZenMLMetadataStore()

    # disallow to/from_config for the base class by checking against
    # factory keys
    with pytest.raises(AssertionError):
        mds1.to_config()


def test_from_config():
    config = {MLMetadataKeys.TYPE: None,
              MLMetadataKeys.ARGS: {}}

    # throws because base MDStore is not in the factory
    with pytest.raises(AssertionError):
        _ = ZenMLMetadataStore.from_config(config)


def test_get_pipeline_status(repo):
    random_pipeline = random.choice(repo.get_pipelines())

    mds1 = ZenMLMetadataStore()

    # TODO: This returns a NotStarted enum, which may be misleading as the
    #  associated store does not even exist
    # with pytest.raises(expected_query_error):
    assert mds1.get_pipeline_status(random_pipeline) == \
           PipelineStatusTypes.NotStarted.name


def test_get_pipeline_executions(repo):
    mds1 = ZenMLMetadataStore()

    random_pipeline = random.choice(repo.get_pipelines())

    # if we query a different metadata store for the pipeline,
    # there should be no executions, i.e. an empty list
    with pytest.raises(expected_query_error):
        _ = mds1.get_pipeline_executions(random_pipeline)


def test_get_components_status(repo):
    mds1 = ZenMLMetadataStore()

    random_pipeline = random.choice(repo.get_pipelines())

    with pytest.raises(expected_query_error):
        _ = mds1.get_components_status(random_pipeline)


def test_get_artifacts_by_component(repo):
    mds1 = ZenMLMetadataStore()

    random_pipeline = random.choice(repo.get_pipelines())

    # pick a component guaranteed to be present
    component_name = GDPComponent.SplitGen.name

    with pytest.raises(expected_query_error):
        _ = mds1.get_artifacts_by_component(random_pipeline,
                                            component_name)


def test_get_component_execution(repo):
    mds1 = ZenMLMetadataStore()

    random_pipeline = random.choice(repo.get_pipelines())

    component_name = GDPComponent.SplitGen.name

    with pytest.raises(expected_query_error):
        _ = mds1.get_component_execution(random_pipeline,
                                         component_name)


def test_get_pipeline_context(repo):
    mds1 = ZenMLMetadataStore()

    random_pipeline = random.choice(repo.get_pipelines())

    with pytest.raises(expected_query_error):
        _ = mds1.get_pipeline_context(random_pipeline)


def test_get_artifacts_by_execution():
    mds1 = ZenMLMetadataStore()

    # no execution possible
    fake_id = "abcdefg"
    with pytest.raises(ValueError):
        _ = mds1.get_artifacts_by_execution(fake_id)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Any, Dict, List, Text

import apache_beam as beam
import tensorflow as tf
from tfx import types
from tfx.dsl.components.base import base_executor
from tfx.types import artifact_utils
from tfx.types.artifact_utils import get_split_uri
from tfx.utils import io_utils

from zenml.components.split_gen.executor import WriteSplit
from zenml.standards.standard_keys import StepKeys
from zenml.steps.split.utils import get_categorical_value
from zenml.steps.tokenizer import BaseTokenizer
from zenml.utils import path_utils, source_utils


def append_tf_example(ex: tf.train.Example, tokenizer_step: BaseTokenizer):
    """
    Append the tokenizer encoding outputs as features to the existing data
    in tf.train.Example format.

    Args:
        ex: tf.train.Example with the raw input features.
        tokenizer_step: Local tokenizer step used in the NLP pipeline.

    Returns:
        A tf.train.Example with all of the features from the `ex` input,
         plus two features `input_ids` and `attention_mask` holding the word
         IDs and attention mask outputs from the tokenizer.

    """
    sentence = get_categorical_value(ex, tokenizer_step.text_feature)

    encoded = tokenizer_step.encode(sentence, output_format="dict")

    new_feature = {k: tf.train.Feature(
        int64_list=tf.train.Int64List(value=v)) for k, v in encoded.items()}

    new_feature.update({
        f: ex.features.feature[f] for f in ex.features.feature
    })

    return tf.train.Example(features=tf.train.Features(feature=new_feature))


class TokenizerExecutor(base_executor.BaseExecutor):
    """
    Tokenizer executor. This component uses a tokenizer, either already
    trained or newly instantiated, and uses it to transform the input data.
    """

    def Do(self, input_dict: Dict[Text, List[types.Artifact]],
           output_dict: Dict[Text, List[types.Artifact]],
           exec_properties: Dict[Text, Any]) -> None:

        source = exec_properties[StepKeys.SOURCE]
        args = exec_properties[StepKeys.ARGS]

        c = source_utils.load_source_path_class(source)
        tokenizer_step: BaseTokenizer = c(**args)

        tokenizer_location = artifact_utils.get_single_uri(
            output_dict["tokenizer"])

        split_uris, split_names, all_files = [], [], []
        for artifact in input_dict["examples"]:
            for split in artifact_utils.decode_split_names(
                    artifact.split_names):
                split_names.append(split)
                uri = os.path.join(artifact.uri, split)
                split_uris.append((split, uri))
                all_files += path_utils.list_dir(uri)

        # Get output split path
        output_examples = artifact_utils.get_single_instance(
            output_dict["output_examples"])
        output_examples.split_names = artifact_utils.encode_split_names(
            split_names)

        if not tokenizer_step.skip_training:
            tokenizer_step.train(files=all_files)

            tokenizer_step.save(output_dir=tokenizer_location)

        with self._make_beam_pipeline() as p:
            for split, uri in split_uris:
                input_uri = io_utils.all_files_pattern(uri)

                _ = (p
                     | 'ReadData.' + split >> beam.io.ReadFromTFRecord(
                            file_pattern=input_uri)
                     | "ParseTFExFromString." + split >> beam.Map(
                            tf.train.Example.FromString)
                     | "AddTokens." + split >> beam.Map(
                            append_tf_example,
                            tokenizer_step=tokenizer_step)
                     | 'Serialize.' + split >> beam.Map(
                            lambda x: x.SerializeToString())
                     | 'WriteSplit.' + split >> WriteSplit(
                            get_split_uri(
                                output_dict["output_examples"],
                                split)))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pipeline to create data sources"""

from typing import Dict, Text, Any, List

from tfx.components.schema_gen.component import SchemaGen
from tfx.components.statistics_gen.component import StatisticsGen

from zenml.components import DataGen
from zenml.enums import GDPComponent
from zenml.pipelines import BasePipeline
from zenml.utils.post_training.post_training_utils import \
    view_statistics, view_schema


class DataPipeline(BasePipeline):
    """DataPipeline definition to create datasources.

    A DataPipeline is used to create datasources in ZenML. Each data pipeline
    creates a snapshot of the datasource in time. All datasources are consumed
    by different ZenML pipelines like the TrainingPipeline.
    """
    PIPELINE_TYPE = 'data'

    def get_tfx_component_list(self, config: Dict[Text, Any]) -> List:
        """
        Creates a data pipeline out of TFX components.

        A data pipeline is used to ingest data from a configured source, e.g.
        local files or cloud storage. In addition, a schema and statistics are
        also computed immediately afterwards for the processed data points.

        Args:
            config: Dict. Contains a ZenML configuration used to build the
             data pipeline.

        Returns:
            A list of TFX components making up the data pipeline.
        """
        data = DataGen(
            name=self.datasource.name,
            source=self.datasource._source,
            source_args=self.datasource._source_args).with_id(
            GDPComponent.DataGen.name
        )
        statistics_data = StatisticsGen(
            examples=data.outputs.examples
        ).with_id(GDPComponent.DataStatistics.name)

        schema_data = SchemaGen(
            statistics=statistics_data.outputs.output,
        ).with_id(GDPComponent.DataSchema.name)

        return [data, statistics_data, schema_data]

    def view_statistics(self, magic: bool = False, port: int = 0):
        """
        View statistics for data pipeline in HTML.

        Args:
            magic (bool): Creates HTML page if False, else
            creates a notebook cell.
            port (int): Port at which to launch the statistics facet.
        """
        uri = self.get_artifacts_uri_by_component(
            GDPComponent.DataStatistics.name)[0]
        view_statistics(uri, magic, port)

    def view_schema(self):
        """View schema of data flowing in pipeline."""
        uri = self.get_artifacts_uri_by_component(
            GDPComponent.DataSchema.name)[0]
        view_schema(uri)

    def steps_completed(self) -> bool:
        return True



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base High-level ZenML Pipeline definition"""

import time
from abc import abstractmethod
from typing import Dict, Text, Any, Optional, List
from uuid import uuid4

from zenml.backends.orchestrator import OrchestratorBaseBackend
from zenml.constants import CONFIG_VERSION
from zenml.datasources import BaseDatasource
from zenml.enums import PipelineStatusTypes
from zenml.exceptions import AlreadyExistsException
from zenml.logger import get_logger
from zenml.metadata import ZenMLMetadataStore
from zenml.repo import Repository, ArtifactStore
from zenml.standards import standard_keys as keys
from zenml.steps import BaseStep
from zenml.utils import source_utils
from zenml.utils.analytics_utils import track, CREATE_PIPELINE, RUN_PIPELINE, \
    GET_PIPELINE_ARTIFACTS
from zenml.utils.print_utils import to_pretty_string, PrintStyles

logger = get_logger(__name__)


class BasePipeline:
    """Base class for all ZenML pipelines.

    Every ZenML pipeline should override this class.
    """
    PIPELINE_TYPE = 'base'

    def __init__(self,
                 name: Text = None,
                 enable_cache: Optional[bool] = True,
                 steps_dict: Dict[Text, BaseStep] = None,
                 backend: OrchestratorBaseBackend = None,
                 metadata_store: Optional[ZenMLMetadataStore] = None,
                 artifact_store: Optional[ArtifactStore] = None,
                 datasource: Optional[BaseDatasource] = None,
                 datasource_commit_id: Optional[Text] = None,
                 pipeline_name: Optional[Text] = None,
                 *args, **kwargs):
        """
        Construct a base pipeline. This is a base interface that is meant
        to be overridden in multiple other pipeline use cases.

        Args:
            name: Outward-facing name of the pipeline.
            pipeline_name: A unique name that identifies the pipeline after
             it is run.
            enable_cache: Boolean, indicates whether or not caching
             should be used.
            steps_dict: Optional dict of steps.
            backend: Orchestrator backend.
            metadata_store: Configured metadata store. If None,
             the default metadata store is used.
            artifact_store: Configured artifact store. If None,
             the default artifact store is used.
            datasource: The datasource to use.
            datasource_commit_id: The datasource commit id to use.
        """
        # Generate a name if not given
        if name is None:
            name = str(round(time.time() * 1000))
        self.name = name
        self._immutable = False

        # Metadata store
        if metadata_store:
            self.metadata_store: ZenMLMetadataStore = metadata_store
        else:
            # use default
            self.metadata_store: ZenMLMetadataStore = \
                Repository.get_instance().get_default_metadata_store()

        if pipeline_name:
            # This means its been loaded in through YAML, try to get context
            self.pipeline_name = pipeline_name
            self.file_name = self.pipeline_name + '.yaml'
        else:
            # if pipeline_name is None then its a new pipeline
            self.pipeline_name = self.create_pipeline_name_from_name()
            self.file_name = self.pipeline_name + '.yaml'
            # check duplicates here as its a 'new' pipeline
            self._check_registered()
            track(event=CREATE_PIPELINE)
            logger.info(f'Pipeline {name} created.')

        self.enable_cache = enable_cache

        if steps_dict is None:
            self.steps_dict: Dict[Text, BaseStep] = {}
        else:
            self.steps_dict = steps_dict

        # Default to local
        if backend is None:
            self.backend = OrchestratorBaseBackend()
        else:
            self.backend = backend

        # Artifact store
        if artifact_store:
            self.artifact_store = artifact_store
        else:
            # use default
            self.artifact_store = \
                Repository.get_instance().get_default_artifact_store()

        # Datasource
        if datasource:
            self.datasource = datasource
            self.datasource_commit_id = datasource.get_latest_commit()
        else:
            self.datasource = None
            self.datasource_commit_id = None

        if datasource_commit_id:
            self.datasource_commit_id = datasource_commit_id

        self._source = source_utils.resolve_class(self.__class__)
        self._kwargs = {
            keys.PipelineDetailKeys.NAME: self.pipeline_name,
            keys.PipelineDetailKeys.ENABLE_CACHE: self.enable_cache,
        }
        if kwargs:
            self._kwargs.update(kwargs)

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    @property
    def is_executed_in_metadata_store(self):
        try:
            # if we find context, then it has been executed
            self.metadata_store.get_pipeline_context(self)
            return True
        except Exception as _:
            return False

    @abstractmethod
    def get_tfx_component_list(self, config: Dict[Text, Any]) -> List:
        """
        Converts config to TFX components list. This is the point in the
        framework where ZenML Steps get translated into TFX pipelines.

        Args:
            config: dict of ZenML config.
        """
        pass

    @abstractmethod
    def steps_completed(self) -> bool:
        """Returns True if all steps complete, else raises exception"""
        pass

    @staticmethod
    def get_name_from_pipeline_name(pipeline_name: Text):
        """
        Gets name from pipeline name.

        Args:
            pipeline_name (str): simple string name.
        """
        return "_".join(pipeline_name.split('_')[1:-1])

    @classmethod
    def from_config(cls, config: Dict):
        """
        Convert from pipeline config to ZenML Pipeline object.

        All steps are also populated and configuration set to parameters set
        in the config file.

        Args:
            config: a ZenML config in dict-form (probably loaded from YAML).
        """
        # start with artifact store
        artifact_store = ArtifactStore(config[keys.GlobalKeys.ARTIFACT_STORE])

        # metadata store
        metadata_store = ZenMLMetadataStore.from_config(
            config=config[keys.GlobalKeys.METADATA_STORE]
        )

        # orchestration backend
        backend = OrchestratorBaseBackend.from_config(
            config[keys.GlobalKeys.BACKEND])

        # pipeline configuration
        p_config = config[keys.GlobalKeys.PIPELINE]
        kwargs = p_config[keys.PipelineKeys.ARGS]
        pipeline_name = kwargs.pop(keys.PipelineDetailKeys.NAME)
        pipeline_source = p_config[keys.PipelineKeys.SOURCE]

        # populate steps
        steps_dict: Dict = {}
        for step_key, step_config in p_config[keys.PipelineKeys.STEPS].items():
            steps_dict[step_key] = BaseStep.from_config(step_config)

        # datasource
        datasource = BaseDatasource.from_config(
            config[keys.GlobalKeys.PIPELINE])

        # datasource commit
        datasource_commit_id = p_config[keys.PipelineKeys.DATASOURCE_COMMIT_ID]

        class_ = source_utils.load_source_path_class(pipeline_source)

        obj = class_(
            steps_dict=steps_dict,
            backend=backend,
            artifact_store=artifact_store,
            metadata_store=metadata_store,
            datasource=datasource,
            datasource_commit_id=datasource_commit_id,
            pipeline_name=pipeline_name,
            name=cls.get_name_from_pipeline_name(pipeline_name),
            **kwargs
        )
        obj._immutable = True
        return obj

    def _check_registered(self):
        if Repository.get_instance().get_pipeline_by_name(
                self.name) is not None:
            raise AlreadyExistsException(
                name=self.name, resource_type='pipeline')

    def _validate_steps(self):
        """Ensure that steps are of the right type"""
        for step_name, step_val in self.steps_dict.items():
            if not issubclass(type(step_val), BaseStep):
                raise AssertionError(
                    f'Step {step_name} needs to be an object that is a '
                    f'sub-class of: {BaseStep}')

    def add_datasource(self, datasource: BaseDatasource,
                       commit_id: Text = None):
        """
        Add datasource to pipeline.

        Args:
            datasource: class of type BaseDatasource.
            commit_id: optionally the commit of the datasource to use. If
            left None, then latest commit is used.
        """
        self.datasource = datasource
        if commit_id is None:
            self.datasource_commit_id = datasource.get_latest_commit()

    def create_pipeline_name_from_name(self):
        """Creates a unique pipeline name from user-provided name."""
        return self.PIPELINE_TYPE.lower() + '_' + self.name + '_' + str(
            uuid4())

    def get_steps_config(self) -> Dict:
        """Convert Step classes to steps config dict."""
        steps_config = {}
        for step_key, step in self.steps_dict.items():
            steps_config[step_key] = step.to_config()

        return {keys.PipelineKeys.STEPS: steps_config}

    def get_pipeline_config(self):
        """Get pipeline config"""
        steps_config = self.get_steps_config()
        steps_config.update({
            keys.PipelineKeys.ARGS: self._kwargs,
            keys.PipelineKeys.SOURCE: self._source,
            keys.PipelineKeys.DATASOURCE: self.datasource.to_config() if
            self.datasource is not None else {},
            keys.PipelineKeys.DATASOURCE_COMMIT_ID: self.datasource_commit_id,
        })
        return steps_config

    def to_config(self) -> Dict:
        """Converts entire pipeline to ZenML config."""
        return {
            keys.GlobalKeys.VERSION: CONFIG_VERSION,
            keys.GlobalKeys.METADATA_STORE: self.metadata_store.to_config(),
            keys.GlobalKeys.ARTIFACT_STORE: self.artifact_store.path,
            keys.GlobalKeys.BACKEND: self.backend.to_config(),
            keys.GlobalKeys.PIPELINE: self.get_pipeline_config(),
        }

    def get_status(self) -> Text:
        """Get status of pipeline."""
        store = self.metadata_store
        return store.get_pipeline_status(self)

    def register_pipeline(self, config: Dict[Text, Any]):
        """
        Registers a pipeline in the artifact store as a YAML file.

        Args:
            config: dict representation of ZenML config.
        """
        self._check_registered()
        Repository.get_instance().register_pipeline(
            file_name=self.file_name, config=config)

    def load_config(self) -> Dict[Text, Any]:
        """Loads a config dict from yaml file."""
        return Repository.get_instance().load_pipeline_config(
            file_name=self.file_name)

    @track(event=GET_PIPELINE_ARTIFACTS)
    def get_artifacts_uri_by_component(
            self, component_name: Text, resolve_locally: bool = True):
        """
        Gets the artifacts of any component within a pipeline. All artifacts
        are resolved locally if resolve_locally flag is set.

        Args:
            component_name (str): name of component
            resolve_locally (bool): If True, artifact is downloadeded locally.
        """
        status = self.metadata_store.get_pipeline_status(self)
        if status != PipelineStatusTypes.Succeeded.name:
            AssertionError('Cannot retrieve as pipeline is not succeeded.')
        artifacts = self.metadata_store.get_artifacts_by_component(
            self, component_name)
        # Download if not base
        uris = []
        for a in artifacts:
            uris.append(self.artifact_store.resolve_uri_locally(
                a.uri) if resolve_locally else a.uri)
        return uris

    def copy(self, new_name: Text):
        """
        Deep copy the pipeline and therefore remove mutability requirement.

        Args:
            new_name (str): New name for copied pipeline.
        """
        class_ = self.__class__
        args = self.__dict__.copy()

        # Doing this will reset immutability
        args.pop('pipeline_name')
        args['name'] = new_name
        obj = class_(**args)
        obj._immutable = False
        return obj

    def run_config(self, config: Dict[Text, Any]):
        """
        Gets TFX pipeline from config.

        Args:
            config: dict of ZenML config.
        """
        assert issubclass(self.backend.__class__, OrchestratorBaseBackend)
        self.backend.run(config)

    @track(event=RUN_PIPELINE)
    def run(self,
            backend: OrchestratorBaseBackend = None,
            metadata_store: Optional[ZenMLMetadataStore] = None,
            artifact_store: Optional[ArtifactStore] = None):
        """
        Run the pipeline associated with the datasource.

        Args:
            backend: orchestrator backend for pipeline.
            metadata_store: chosen metadata store, if None use default
            artifact_store: chosen artifact store, if None use default
             """

        # TODO: [HIGH] Important, think about separating register and run
        #  and that way ask user to input name while registering pipeline.

        # Resolve default
        if metadata_store:
            logger.warning('Changing the metadata_store or artifact_store '
                           'might cause your pipelines to be '
                           'non-reproducible and non-comparable.')
            self.metadata_store = metadata_store

        if artifact_store:
            logger.warning('Changing the metadata_store or artifact_store '
                           'might cause your pipelines to be '
                           'non-reproducible and non-comparable.')
            self.artifact_store = artifact_store

        if backend:
            self.backend = backend

        # We do not allow ml metadata to get polluted by repeated runs
        if self.is_executed_in_metadata_store:
            logger.info(f'Pipeline: `{self.name}` has already been executed '
                        f'for the connected metadata store. Pipelines are '
                        f'immutable after they run within the same metadata '
                        f'store. If the pipeline failed, then try using '
                        f'the `pipeline.copy(new_name)` method to copy the '
                        f'current `pipeline` and give it a new name.')
            return

        # Check if steps are complete
        self.steps_completed()

        # Validate steps
        self._validate_steps()

        # assert some datasource stuff
        if self.datasource and self.datasource.metadata_store.to_config() != \
                self.metadata_store.to_config():
            raise AssertionError(
                'The metadata_store of the specified datasource is different '
                'from this pipeline. Pipelines and their datasources need to '
                'be in the same metadata_store.')
        if self.datasource and self.datasource.artifact_store.path != \
                self.artifact_store.path:
            raise AssertionError(
                'The artifact_store of the specified datasource is different '
                'from this pipeline. Pipelines and their datasources need to '
                'be in the same artifact_store.')

        if self._immutable:
            # This means its an 'older' pipeline that has been loaded in via
            # YAML but does not exist in the metadata store, so we are safe
            # to run it without inferring new version pins. This will result
            # in a repeated run.
            logger.info('Pipeline is an immutable state: Any changes made '
                        'after loading it will not be reflected in this run.')
            config = self.load_config()
        else:
            # This means its a brand new pipeline
            config = self.to_config()
            # Register in the repository
            self.register_pipeline(config)

        self.run_config(config)

        # After running, pipeline is immutable
        self._immutable = True



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import apache_beam as beam
from tfx.orchestration import pipeline
from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner, \
    _ComponentAsDoFn
from tfx.orchestration.config import config_utils

from zenml.logger import get_logger

logger = get_logger(__name__)


class ZenMLBeamlDagRunner(BeamDagRunner):
    """
    This is the same as the super class from tfx:
    tfx.orchestration.beam.beam_dag_runner.BeamDagRunner with the exception
    being that the pipeline_run is not overridden. Full credit to Google LLC
    for the original source code found at:
    https://github.com/tensorflow/tfx/tree/master/tfx/orchestration/beam/
    """

    def run(self, tfx_pipeline: pipeline.Pipeline) -> None:

        with beam.Pipeline(argv=self._beam_orchestrator_args) as p:
            root = p | 'CreateRoot' >> beam.Create([None])

            signal_map = {}
            for component in tfx_pipeline.components:
                component_id = component.id

                signals_to_wait = []
                if component.upstream_nodes:
                    for upstream_node in component.upstream_nodes:
                        assert upstream_node in signal_map, (
                            'Components is not in '
                            'topological order')
                        signals_to_wait.append(
                            signal_map[upstream_node])
                logger.debug('Component %s depends on %s.',
                             component_id,
                             [s.producer.full_label for s in
                              signals_to_wait])

                (component_launcher_class,
                 component_config) = \
                    config_utils.find_component_launch_info(
                        self._config, component)

                signal_map[component] = (
                        root
                        | 'Run[%s]' % component_id >> beam.ParDo(
                    _ComponentAsDoFn(component,
                                     component_launcher_class,
                                     component_config, tfx_pipeline),
                    *[beam.pvalue.AsIter(s) for s in signals_to_wait]))
                logger.debug('Component %s is scheduled.',
                             component_id)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text, List

from zenml.steps import BaseStep


class BaseTokenizer(BaseStep):
    """
    Base step class for all tokenizers.
    """

    def __init__(self,
                 text_feature: Text,
                 skip_training: bool = False,
                 **kwargs):
        super(BaseTokenizer, self).__init__(text_feature=text_feature,
                                            skip_training=skip_training,
                                            **kwargs)

        self.text_feature = text_feature
        self.skip_training = skip_training

    def train(self, files: List[Text]):
        raise NotImplementedError

    def encode(self, sequence: Text, output_format: Text):
        raise NotImplementedError

    def save(self, output_dir: Text):
        raise NotImplementedError

    def load_vocab(self, path_to_vocab: Text):
        raise NotImplementedError



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import List, Text, Dict

import tensorflow_data_validation as tfdv
from tfx import types
from tfx.components.schema_gen.executor import _DEFAULT_FILE_NAME, SCHEMA_KEY
from tfx.types import artifact_utils
from tfx.types.artifact import Artifact
from tfx.utils import io_utils


def parse_statistics(split_name: Text,
                     statistics: List[Artifact]) -> Dict[Text, int]:
    stats_uri = io_utils.get_only_uri_in_dir(
        artifact_utils.get_split_uri(statistics, split_name))

    stats = tfdv.load_statistics(stats_uri)

    return stats


def parse_schema(input_dict: Dict[Text, List[types.Artifact]]):
    schema = input_dict.get(SCHEMA_KEY, None)
    if not schema:
        return schema
    else:
        schema_path = os.path.join(
            artifact_utils.get_single_uri(schema),
            _DEFAULT_FILE_NAME)
        schema_reader = io_utils.SchemaReader()
        parsed_schema = schema_reader.read(schema_path)

        return parsed_schema



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text, Dict, Any

import apache_beam as beam
from tfx.components.example_gen.base_example_gen_executor import _WriteSplit

from zenml.components.data_gen import utils


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.typehints.Dict[Text, Any])
@beam.typehints.with_output_types(beam.pvalue.PDone)
def WriteToTFRecord(datapoints: Dict[Text, Any],
                    schema: Dict[Text, Any],
                    output_split_path: Text) -> beam.pvalue.PDone:
    """Infers schema and writes to TFRecord"""
    # Obtain the schema
    if schema:
        schema_dict = {k: utils.SCHEMA_MAPPING[v] for k, v in schema.items()}
    else:
        schema = (datapoints
                  | 'Schema' >> beam.CombineGlobally(utils.DtypeInferrer()))
        schema_dict = beam.pvalue.AsSingleton(schema)

    return (datapoints
            | 'ToTFExample' >> beam.Map(utils.append_tf_example, schema_dict)
            | 'WriteToTFRecord' >> _WriteSplit(
                output_split_path=output_split_path
            ))



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.backends import BaseBackend


def test_to_from_config(equal_backends):

    kwargs = {"a": 1, "name": "my_backend", "grade": 1.5}
    b1 = BaseBackend(**kwargs)

    b2 = BaseBackend.from_config(b1.to_config())

    assert equal_backends(b1, b2, loaded=True)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the base Orchestrator Backend"""

import os
from typing import Dict, Text, Any

from tfx.orchestration import pipeline

from zenml.backends import BaseBackend
from zenml.backends.orchestrator.base.zenml_local_orchestrator import \
    ZenMLLocalDagRunner
from zenml.backends.processing import ProcessingBaseBackend


class OrchestratorBaseBackend(BaseBackend):
    """
    Local ZenML orchestrator backend. Use this to run a ZenML pipeline
    locally on a machine.

    An orchestrator backend is responsible for scheduling, initializing and
    running different pipeline components. Examples of orchestrators are
    Apache Beam, Kubeflow or (here) Local Orchestration.

    Abstracting the pipeline logic from the orchestrator backend enables
    machine learning workloads to be run in different kinds of environments.
    For larger, decentralized data processing applications, a cloud-based
    backend can be used to distribute work across multiple machines.
    For quick prototyping and local tests, a single-machine direct backend can
    be selected to execute an ML Pipeline with minimal orchestration overhead.
    """
    BACKEND_TYPE = 'orchestrator'

    @staticmethod
    def get_tfx_pipeline(config: Dict[Text, Any]) -> pipeline.Pipeline:
        """
        Converts ZenML config dict to TFX pipeline.

        Args:
            config: A ZenML config dict

        Returns:
            tfx_pipeline: A TFX pipeline object.
        """
        from zenml.pipelines import BasePipeline
        zen_pipeline: BasePipeline = BasePipeline.from_config(config)

        # Get component list
        component_list = zen_pipeline.get_tfx_component_list(config)

        # Get pipeline metadata
        pipeline_name = zen_pipeline.pipeline_name
        metadata_connection_config = \
            zen_pipeline.metadata_store.get_tfx_metadata_config()
        artifact_store = zen_pipeline.artifact_store

        # Pipeline settings
        pipeline_root = os.path.join(
            artifact_store.path, artifact_store.unique_id)
        pipeline_log = os.path.join(pipeline_root, 'logs', pipeline_name)

        # Resolve execution backend
        execution = ProcessingBaseBackend()  # default
        for e in zen_pipeline.steps_dict.values():
            # find out the processing backends, take the first one which is
            # not a ProcessingBaseBackend
            if e.backend and issubclass(
                    e.backend.__class__, ProcessingBaseBackend) and \
                    e.backend.__class__ != ProcessingBaseBackend:
                execution = e.backend
                break

        beam_args = execution.get_beam_args(pipeline_name, pipeline_root)

        tfx_pipeline = pipeline.Pipeline(
            components=component_list,
            beam_pipeline_args=beam_args,
            metadata_connection_config=metadata_connection_config,
            pipeline_name=zen_pipeline.artifact_store.unique_id,  # for caching
            pipeline_root=pipeline_root,
            log_root=pipeline_log,
            enable_cache=zen_pipeline.enable_cache)

        # Ensure that the run_id is ZenML pipeline_name
        tfx_pipeline.pipeline_info.run_id = zen_pipeline.pipeline_name
        return tfx_pipeline

    def run(self, config: Dict[Text, Any]):
        """
        This run function essentially calls an underlying TFX orchestrator run.
        However it is meant as a higher level abstraction with some
        opinionated decisions taken.

        Args:
            config: a ZenML config dict
        """
        tfx_pipeline = self.get_tfx_pipeline(config)
        ZenMLLocalDagRunner().run(tfx_pipeline)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Cortex pusher executor."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import tempfile
from typing import Any, Dict, List, Text

import cortex
from tfx import types
from tfx.components.pusher import executor as tfx_pusher_executor
from tfx.types import artifact_utils
from tfx.utils import io_utils
from tfx.utils import json_utils
from tfx.utils import path_utils

# Keys for custom_config.
_CUSTOM_CONFIG_KEY = 'custom_config'
SERVING_ARGS_KEY = 'cortex_serving_args'


class Executor(tfx_pusher_executor.Executor):
    """Deploy a model to Cortex serving."""

    def Do(self, input_dict: Dict[Text, List[types.Artifact]],
           output_dict: Dict[Text, List[types.Artifact]],
           exec_properties: Dict[Text, Any]):
        """Overrides the tfx_pusher_executor.

        Args:
          input_dict: Input dict from input key to a list of artifacts,
          including:
            - model_export: exported model from trainer.
            - model_blessing: model blessing path from evaluator.
          output_dict: Output dict from key to a list of artifacts, including:
            - model_push: A list of 'ModelPushPath' artifact of size one. It
            will
              include the model in this push execution if the model was pushed.
          exec_properties: Mostly a passthrough input dict for
            tfx.components.Pusher.executor.custom_config
        Raises:
          ValueError: if custom config not present or not a dict.
          RuntimeError: if
        """
        self._log_startup(input_dict, output_dict, exec_properties)

        # check model blessing
        model_push = artifact_utils.get_single_instance(
            output_dict[tfx_pusher_executor.PUSHED_MODEL_KEY])
        if not self.CheckBlessing(input_dict):
            self._MarkNotPushed(model_push)
            return

        model_export = artifact_utils.get_single_instance(
            input_dict[tfx_pusher_executor.MODEL_KEY])

        custom_config = json_utils.loads(
            exec_properties.get(_CUSTOM_CONFIG_KEY, 'null'))
        if custom_config is not None and not isinstance(custom_config, Dict):
            raise ValueError(
                'custom_config in execution properties needs to be a '
                'dict.')

        cortex_serving_args = custom_config.get(SERVING_ARGS_KEY)
        if not cortex_serving_args:
            raise ValueError(
                '\'cortex_serving_args\' is missing in \'custom_config\'')

        # Deploy the model.
        io_utils.copy_dir(
            src=path_utils.serving_model_path(model_export.uri),
            dst=model_push.uri)
        model_path = model_push.uri

        # Cortex implementation starts here
        # pop the env and initialize client
        cx = cortex.client(cortex_serving_args.pop('env'))

        # load the predictor
        predictor_path = cortex_serving_args.pop('predictor_path')
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            temp_project_dir = tmp_dir_name

            # predictor
            p_dump_path = os.path.join(temp_project_dir, 'predictor.py')
            io_utils.copy_file(predictor_path, p_dump_path)

            # requirements.txt
            reqs = cortex_serving_args.pop('requirements', [])
            if reqs :
                r_dump_path = os.path.join(
                    temp_project_dir, 'requirements.txt')
                io_utils.write_string_file(r_dump_path, '\n'.join(reqs))

            # conda-packages.txt
            c_reqs = cortex_serving_args.pop('conda_packages', [])
            if c_reqs:
                r_dump_path = os.path.join(
                    temp_project_dir, 'conda-packages.txt')
                io_utils.write_string_file(r_dump_path, '\n'.join(c_reqs))

            # edit the api_config
            api_config = cortex_serving_args.pop('api_config')
            if 'config' not in api_config['predictor']:
                api_config['predictor']['config'] = {}
            api_config['predictor']['config']['model_artifact'] = model_path

            # launch the api
            api_config['predictor']['path'] = 'predictor.py'

            # configure the model path
            if 'models' not in api_config['predictor']:
                api_config['predictor']['models'] = {}
            api_config['predictor']['models'].update({'path': model_path})
            cx.create_api(
                api_config,
                project_dir=temp_project_dir,
                **cortex_serving_args)

        self._MarkPushed(
            model_push,
            pushed_destination=model_path)



from zenml.datasources import CSVDatasource
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository
from zenml.steps.evaluator import TFMAEvaluator
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.exceptions import AlreadyExistsException

from zenml.utils.naming_utils import transformed_label_name

# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = CSVDatasource(name='Pima Indians Diabetes',
                       path='gs://zenml_quickstart/diabetes.csv')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'Pima Indians Diabetes')
training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    StandardPreprocesser(
        features=['times_pregnant', 'pgc', 'dbp', 'tst', 'insulin', 'bmi',
                  'pedigree', 'age'],
        labels=['has_diabetes'],
        overwrite={'has_diabetes': {
            'transform': [{'method': 'no_transform', 'parameters': {}}]}}
    ))

# Add a trainer
training_pipeline.add_trainer(TFFeedForwardTrainer(
    loss='binary_crossentropy',
    last_activation='sigmoid',
    output_units=1,
    metrics=['accuracy'],
    epochs=20))

# Add an evaluator
training_pipeline.add_evaluator(
    TFMAEvaluator(slices=[['has_diabetes']],
                  metrics={transformed_label_name('has_diabetes'):
                               ['binary_crossentropy', 'binary_accuracy']}))

# Run the pipeline locally
training_pipeline.run()

# See schema of data
training_pipeline.view_schema()

# See statistics of train and eval
training_pipeline.view_statistics()

# Creates a notebook for evaluation
training_pipeline.evaluate()



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text, Dict, Any, Optional

from tfx.dsl.components.base.base_component import BaseComponent
from tfx.dsl.components.base.executor_spec import ExecutorClassSpec
from tfx.types import standard_artifacts, Channel
from tfx.types.component_spec import ComponentSpec, ExecutionParameter, \
    ChannelParameter

from zenml.components.tokenizer.constants import EXAMPLES, TOKENIZER, \
    OUTPUT_EXAMPLES
from zenml.components.tokenizer.executor import TokenizerExecutor
from zenml.standards.standard_keys import StepKeys


class TokenizerSpec(ComponentSpec):
    PARAMETERS = {
        StepKeys.SOURCE: ExecutionParameter(type=Text),
        StepKeys.ARGS: ExecutionParameter(type=Dict[Text, Any]),
    }
    INPUTS = {
        EXAMPLES: ChannelParameter(type=standard_artifacts.Examples),
    }

    OUTPUTS = {
        TOKENIZER: ChannelParameter(type=standard_artifacts.Model),
        OUTPUT_EXAMPLES: ChannelParameter(type=standard_artifacts.Examples),
    }


class Tokenizer(BaseComponent):
    SPEC_CLASS = TokenizerSpec
    EXECUTOR_SPEC = ExecutorClassSpec(TokenizerExecutor)

    def __init__(self,
                 source: Text,
                 source_args: Dict[Text, Any],
                 examples: Optional[ChannelParameter] = None,
                 tokenizer: Optional[ChannelParameter] = None,
                 output_examples: Optional[ChannelParameter] = None,
                 instance_name: Optional[Text] = None
                 ):
        """
        Interface for the Tokenizer component. This handles the vocabulary
        construction as well as encoding input sequences for the NLP model.

        Args:
            source: Tokenizer step source class.
            source_args: Keyword arguments for tokenizer step source class.
            examples: Input channel for incoming data.
            tokenizer: Output channel for the tokenizer output.
            output_examples: Output channel for tokenized examples.
            instance_name: Name of the component.
        """
        examples = examples or Channel(type=standard_artifacts.Examples)
        tokenizer = tokenizer or Channel(type=standard_artifacts.Model)
        output_examples = output_examples or Channel(
            type=standard_artifacts.Examples)

        # Initiate the spec and create instance
        spec = self.SPEC_CLASS(source=source,
                               args=source_args,
                               tokenizer=tokenizer,
                               examples=examples,
                               output_examples=output_examples)

        super(Tokenizer, self).__init__(spec=spec,
                                        instance_name=instance_name)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.steps.trainer import BaseTrainerStep
from zenml.utils import requirement_utils

requirement_utils.check_integration(requirement_utils.PYTORCH_INTEGRATION)


class TorchBaseTrainerStep(BaseTrainerStep):
    """
    Base class for all PyTorch based trainer steps. All pytorch based
    trainings should use this as the base class. An example is available
    with torch_ff_trainer.FeedForwardTrainer.
    """
    pass



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.steps.tokenizer.base_tokenizer import BaseTokenizer
from zenml.logger import get_logger
from zenml.utils.requirement_utils import check_integration, \
    HUGGINGFACE_INTEGRATION

logger = get_logger(__name__)

# HuggingFace extra requirement
try:
    check_integration(HUGGINGFACE_INTEGRATION)
    from zenml.steps.tokenizer.hf_tokenizer import \
        HuggingFaceTokenizerStep
except ModuleNotFoundError as e:
    logger.debug(f"There were failed imports due to missing integrations. "
                 f"HuggingFaceTokenizerStep was not imported. "
                 f"More information:")
    logger.debug(e)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text

import click
from tabulate import tabulate

from zenml.cli.cli import cli
from zenml.cli.utils import pretty_print, pass_repo
from zenml.repo import Repository


@cli.group()
def datasource():
    """Data source group"""
    pass


@datasource.command("list")
@pass_repo
def list_datasources(repo: Repository):
    datasources = repo.get_datasources()

    click.echo(tabulate([ds.to_config() for ds in datasources],
                        headers="keys"))


@datasource.command("get")
@click.argument('datasource_name')
@pass_repo
def get_datasource_by_name(repo: Repository, datasource_name: Text):
    """
    Gets pipeline from current repository by matching a name identifier
    against the data source name.

    """
    pretty_print(repo.get_datasource_by_name(datasource_name))
